import imageio
from pylab import *
from qutip import *
from tqdm import tqdm


def animate_bloch(states, duration=0.1, save_all=False, filename: str = "default"):
    b = Bloch()
    b.vector_color = ["r"]
    b.view = [-40, 30]

    images = []
    try:
        length = len(states)
    except:
        length = 1
        states = [states]

    ## normalize colors to the length of data ##
    nrm = mpl.colors.Normalize(0, length)
    colors = cm.cool(nrm(range(length)))  # options: cool, summer, winter, autumn etc.

    ## customize sphere properties ##
    b.point_color = list(colors)  # options: 'r', 'g', 'b' etc.
    b.point_marker = ["o"]
    b.point_size = [30]

    for i in tqdm(
        range(length), desc="Animate thetas on bloch sphere", unit="iteration"
    ):
        b.clear()
        b.add_states(states[i])
        b.add_states(states[: (i + 1)], "point")

        if save_all:
            b.save(dirc="tmp")  # saving images to tmp directory
            temp_filename = "tmp/bloch_%01d.png" % i
        else:
            filename_ = "file_outputs/temp_file.png"
            b.save(temp_filename)

        images.append(imageio.imread(temp_filename))

    imageio.mimsave(f"{temp_filename}.gif", images, duration=duration)



import matplotlib.pyplot as plt
import numpy as np
import requests

mss_url = "http://qtl-axean.mc2.chalmers.se:8002"


def plot_hist(ax: object, job_id: str):
    response = requests.get(mss_url + "/rng/" + job_id)
    if not response.ok:
        raise RuntimeError(
            f"Unable to communicate with {mss_url}, response: {response}"
        )

    data = response.json()
    numbers = data["numbers"]
    # N = data["N"]
    return ax.hist(numbers)


# returns 422 if cant find job_id, which is not response.ok
job_ids = [
    "5a00247d-d22b-44ac-a00a-5d21c087f65d",
    "e904e367-b64b-4380-92eb-151f06ed5faf",
    "d044505c-4c92-4a34-8f7a-b880a03d8069",
]

fig, ax = plt.subplots(len(job_ids), 1, figsize=(5, 10))

for idx, j in enumerate(job_ids):
    ax[idx].set_title(f"job_id: {j}")
    plot_hist(ax=ax[idx], job_id=j)

fig.tight_layout()
plt.show()



import time
from uuid import uuid4 as uuid

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import requests
from qiskit.providers.jobstatus import JobStatus
from qiskit.providers.tergite import Tergite
from tqdm.auto import tqdm

# Task: Estimate Pi numerically
# Area of circle = pi*r**2
#
# If we compute the area of a quarter of a circle and multiply it by 4, we should get Pi = 3.14159...
# The size of the circle doesn't matter, so let's take the top right quadrant of the unit circle.

# ---------------------- settings
vm_url = "http://qtl-axean.mc2.chalmers.se"
mss_url = vm_url + ":8002"
bcc_url = vm_url + ":8000"
chalmers = Tergite.get_provider()
backend = chalmers.get_backend("Nov7")
print(f"Connected to backend: {backend.name}")

text_settings = {"fontsize": 20}
plt.ion()  # enable interactive mode
fig, axs = plt.subplots(1, 2, figsize=(14, 6.5), sharex=True, sharey=True)
# ---------------------- settings


def f(x):
    return np.sqrt(1 - x**2)


def mc_hitmiss(fn: callable, *, M: int) -> tuple:
    np.random.seed(0)
    U = np.random.rand(M, 2)
    estimate = np.mean(
        U[:, 1] <= fn(U[:, 0])
    )  # <- This is the estimator for integral value
    return estimate, U


def plot_estimate(ax: object, estimate: float):
    textstr = "$\hat\pi = {:10.5f}$".format(estimate)
    props = dict(boxstyle="round", facecolor="wheat", alpha=1)
    # place a text box in upper left in axes coords
    ax.text(
        0.15,
        0.15,
        textstr,
        transform=ax.transAxes,
        verticalalignment="top",
        bbox=props,
        fontsize=25,
    )


def plot_samples(ax: object, samples: np.array, marker: str = "o"):
    inside = np.sqrt(samples[:, 0] ** 2 + samples[:, 1] ** 2) < 1
    ax.scatter(samples[inside, 0], samples[inside, 1], color="C1", marker=marker)
    ax.scatter(samples[~inside, 0], samples[~inside, 1], color="C2", marker=marker)


def api_error(url, response):
    return f"Failed to communicate with {url}, response: {response}"


def wait_for_job(job_id: str, timeout: int = 20) -> object:
    mss_response = requests.get(mss_url + "/rng/" + job_id)
    return mss_response


def qmc_estimate(fn: callable, U: np.ndarray) -> float:
    estimate = np.mean(
        U[:, 1] <= fn(U[:, 0])
    )  # <- This is the estimator for integral value
    return estimate


def qmc_hitmiss_parts(fn: callable, num_jobs: int = 3) -> tuple:
    job_ids = list(str(uuid()) for _ in range(num_jobs))

    estimates = list()  # should be monotonic increasing
    points = np.asarray([[], []]).T
    points_by_job = dict()

    for job_id in tqdm(job_ids):
        # queue job on backend directly
        bcc_response = requests.get(bcc_url + "/rng/" + job_id)
        if not bcc_response.ok:
            raise RuntimeError(api_error(bcc_url, bcc_response))

        # wait for job to complete
        mss_response = wait_for_job(job_id)
        data = mss_response.json()

        # split obtained numbers into two disjoint subsets
        N = data["N"] // 2
        X = np.asarray(data["numbers"][:N]).astype(float)  # first M are X coordinate
        Y = np.asarray(data["numbers"][N : (N + N)]).astype(
            float
        )  # second M are Y coordinate

        # these numbers are random 32-bit integers in the range [-2**31, 2**31 - 1]
        # so we want to convert them to floats in the range [0,1]
        X += 2**31
        Y += 2**31
        X *= 2**-32
        Y *= 2**-32

        # concatenate points obtained
        U = np.zeros((N, 2))
        U[:, 0] = X
        U[:, 1] = Y
        points = np.concatenate((points, U))
        points_by_job[job_id] = U
        est = qmc_estimate(fn, points)

        plot_samples(axs[1], U)
        plot_estimate(axs[1], est * 4)
        fig.canvas.draw()
        fig.canvas.flush_events()

        # compute estimator on all the points, up until now
        estimates.append(est)

    return estimates, points_by_job, job_ids


xdef = np.linspace(0, 1)

# ---------- Riemann estimate
riemann_est = sum(f(xdef) * np.diff(xdef)[0])
riemann_est_pi = riemann_est * 4

# ---------- Monte Carlo estimate, from pseudo random numbers
mc_est, mc_samples = mc_hitmiss(f, M=400)
mc_est_pi = mc_est * 4

# ---------- Plotting non quantum stuff
# plot riemann rectangles
riemann_rects = [
    mpl.patches.Rectangle(
        xy=(x, 0),
        width=np.diff(xdef)[0],
        height=f(x),
        ec="black",
        fc="none",
        alpha=0.75,
    )
    for x in xdef
]
for r in riemann_rects:
    axs[0].add_patch(r)
plot_estimate(axs[0], riemann_est_pi)

axs[0].set_title("Riemann", **text_settings)
axs[1].set_title("Quantum Monte Carlo", **text_settings)

for ax in axs:
    ax.plot(np.linspace(0, 1, 1000), f(np.linspace(0, 1, 1000)), lw=3.5, zorder=99)
    ax.set_ylabel("f(x)", **text_settings)
    ax.set_xlabel("x", **text_settings)
    ax.grid(zorder=-99, alpha=0.5)
    ax.tick_params(axis="both", which="minor", labelsize=text_settings["fontsize"])
    ax.tick_params(axis="both", which="major", labelsize=text_settings["fontsize"])
fig.suptitle("$f(x) = (1-x^2)^{1/2}$", **text_settings)

# plot_samples(axs[1], mc_samples)
# plot_estimate(axs[1], mc_est_pi)

fig.canvas.draw()
fig.canvas.flush_events()

# ---------- Plotting quantum stuff
qmc_hitmiss_parts(f, num_jobs=5)

fig.tight_layout()



from qiskit import *

from tergite.qiskit.providers import Tergite

provider = Tergite.get_provider()
pingu_backend = provider.get_backend("pingu")


quantum_circuit = QuantumCircuit.from_qasm_file("tests/example.qasm")
print(quantum_circuit)



import typing

import numpy as np
import qiskit
import torch
import torch.nn as nn
import torch.nn.functional as F
from circuits import HybridLayerCircuit
from torch import optim
from torch.autograd import Function
from tqdm import tqdm


class HybridFunction(Function):
    """Hybrid quantum - classical function definition"""

    @staticmethod
    def forward(ctx, input, quantum_circuit, shift):
        """Forward pass computation"""
        ctx.shift = shift
        ctx.quantum_circuit = quantum_circuit

        expectation_z = ctx.quantum_circuit.run(input[0].tolist())
        result = torch.tensor([expectation_z])
        ctx.save_for_backward(input, result)

        return result

    @staticmethod
    def backward(ctx, grad_output):
        """Backward pass computation"""
        input, expectation_z = ctx.saved_tensors
        input_list = np.array(input.tolist())

        shift_right = input_list + np.ones(input_list.shape) * ctx.shift
        shift_left = input_list - np.ones(input_list.shape) * ctx.shift

        gradients = []
        for i in range(len(input_list)):
            expectation_right = ctx.quantum_circuit.run(shift_right[i])
            expectation_left = ctx.quantum_circuit.run(shift_left[i])

            gradient = torch.tensor([expectation_right]) - torch.tensor(
                [expectation_left]
            )
            gradients.append(gradient)
        # gradients = np.array([gradients]).T
        # gradients = torch.t(torch.stack(gradients))
        gradients = torch.stack(gradients)
        return gradients.double() * grad_output.double(), None, None


class Hybrid(nn.Module):
    """Hybrid quantum - classical layer definition"""

    def __init__(
        self,
        backend: qiskit.providers.Backend,
        shots: int,
        shift,
        qubits: typing.List[int] = None,
        run_id: str = "default",
    ):
        super(Hybrid, self).__init__()
        self.quantum_circuit = HybridLayerCircuit(
            backend=backend, shots=shots, qubits=qubits, run_id=run_id
        )
        self.shift = shift
        self.run_id = run_id

    def forward(self, input):
        return HybridFunction.apply(input, self.quantum_circuit, self.shift)


class Net(nn.Module):
    def __init__(
        self,
        run_id,
        n_qubits: int = 1,
        qubits: typing.List[int] = None,
        backend=qiskit.Aer.get_backend("aer_simulator"),
        shots: int = 1024,
        n_neurons: int = 16,
    ):
        super(Net, self).__init__()

        # TODO: We could fine tune here, if we want to have qubits bins and more combinations of layers
        # in_hybrid_layer = n_qubits if qubits is None else len(qubits)
        in_hybrid_layer = 1

        self.run_id = run_id

        self.fc1 = nn.Linear(2, n_neurons)
        self.fc2 = nn.Linear(n_neurons, in_hybrid_layer)

        self.hybrid = Hybrid(backend, shots, np.pi / 2, qubits=qubits, run_id=run_id)

    def forward(self, x):
        x = x.view(1, -1)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        x = self.hybrid(x)
        return torch.cat((x, 1 - x), -1)


def train_model(
    model: nn.Module, train_loader, epochs: int, plot: bool = False
) -> typing.Tuple["nn.Module", typing.List[float]]:
    # TODO: If one would like to write a beautiful ML framework, it might be better to pass the optimizer as a parameter
    optimizer = optim.Adam(model.parameters(), lr=0.001)
    loss_func = nn.NLLLoss()

    loss_list = []

    model.train()
    for epoch in range(epochs):
        total_loss = []
        for batch_idx, (data, target) in tqdm(
            enumerate(train_loader), desc=f"{epoch = }", unit="samples"
        ):
            optimizer.zero_grad()
            output = model(data)
            loss = loss_func(output, target)
            loss.backward()
            optimizer.step()

            total_loss.append(loss.item())

        loss_list.append(sum(total_loss) / len(total_loss))

        print(
            "\nTraining [{:.0f}%]\tLoss: {:.4f}".format(
                100.0 * (epoch + 1) / epochs, loss_list[-1]
            )
        )

    return model, loss_list


def eval_model(model: nn.Module, eval_loader):
    model.eval()
    loss_func = nn.NLLLoss()
    total_loss = []
    with torch.no_grad():
        correct = 0
        for batch_idx, (data, target) in enumerate(eval_loader):
            output = model(data)

            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()

            loss = loss_func(output, target)
            total_loss.append(loss.item())

        accuracy = sum(total_loss) / len(total_loss)
        print(
            "Performance on test data:\n\tLoss: {:.4f}\n\tAccuracy: {:.1f}%".format(
                accuracy, correct / len(eval_loader) * 100
            )
        )

    return accuracy



import contextlib
import time

import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
import qiskit.circuit as circuit
import qiskit.compiler as compiler
from qiskit.ignis.verification.tomography import (
    StateTomographyFitter,
    state_tomography_circuits,
)
from qiskit.providers.jobstatus import JobStatus
from qiskit.providers.tergite import Tergite
from qiskit.visualization import plot_bloch_multivector
from tqdm.auto import tqdm

chalmers = Tergite.get_provider()
backend = chalmers.get_backend("Nov7")
backend.set_options(shots=2000)

print(f"Loaded backend {backend.name}")

fig, ax = plt.subplots()

# show first frame
im = ax.imshow(plt.imread("ground_state.jpg"))
ax.axis("off")

thetadef = -1 * np.asarray([0, np.pi / 2, np.pi])


def tomog_circs(theta):
    q = circuit.QuantumRegister(1)
    circ = circuit.QuantumCircuit(q)
    circ.barrier([0])
    circ.reset([0])

    circ.rx(theta, q[0])

    return state_tomography_circuits(circ, [q[0]])


print("Transpiling...")
with contextlib.redirect_stderr(None):
    precomputed_tomog_circs = [
        compiler.transpile(tomog_circs(theta), backend=backend) for theta in thetadef
    ]


# function to update figure
def updatefig(j):
    if not j:
        return [im]

    print()
    job = backend.run(precomputed_tomog_circs[j], meas_level=2, meas_return="single")
    print("Computing frame...")
    while job.status() != JobStatus.DONE:
        print("#", end="")
        time.sleep(1)  # blocking wait
    print("  Complete")

    # fit state vector
    fitter = StateTomographyFitter(job.result(), precomputed_tomog_circs[j])

    density_matrix = fitter.fit(method="lstsq")

    # plot and save to main memory
    _tmp = plot_bloch_multivector(
        density_matrix, reverse_bits=True, filename=f"frames/frame{j}.jpg"
    )
    plt.close(_tmp)

    # load file (workaround, because cannot access the axis directly)
    image_j = plt.imread(f"frames/frame{j}.jpg")

    # show next frame
    im.set_array(image_j)
    return [im]


# start animation
ani = animation.FuncAnimation(
    fig, updatefig, frames=range(len(thetadef)), interval=50, blit=True, repeat=False
)
# figManager = plt.get_current_fig_manager()
# figManager.window.showMaximized()
plt.show()



#!/usr/bin/env python3.9
# coding: utf-8

import time
from datetime import datetime
from os import listdir, makedirs
from pathlib import Path
from shutil import move

import numpy as np
from PIL import Image

folder = Path("testdata").resolve()


def save_old_frames():
    global folder
    saved_folder = folder / "saved_animations"
    makedirs(saved_folder, exist_ok=True)

    old_frames = [f for f in listdir(folder) if f.endswith(".jpg")]

    if len(old_frames):
        now_time = datetime.now()
        mstr = now_time.strftime("%Y%m%d%H%M%S")
        new_dir = saved_folder / mstr
        makedirs(new_dir, exist_ok=True)
        for f in old_frames:
            move(folder / f, new_dir / f)


save_old_frames()


def compute_new_frame(j: int):
    # Do some labour, which computes a frame and saves
    # to a folder...
    noise = np.random.rand(50, 50)
    img = Image.fromarray(noise, "RGB")
    img.save(folder / f"frame{j}.jpg")

    # emulate time it actually takes to compute frame
    time.sleep(np.random.rand() * 5)


[compute_new_frame(j) for j in range(20) if not print("#", end="", flush=True)]
print()



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020, 2021
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


from .types.backend import OpenPulseBackend, OpenQASMBackend
from .types.job import Job
from .types.provider import Provider
from .types.provider_factory import ProviderFactory

Tergite = ProviderFactory()



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
# (C) Copyright Stefan Hill 2023
# (C) Copyright Martin Ahindura 2023, 2024
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# This code was refactored from the original on 22nd September, 2023 by Martin Ahindura
"""A sample script doing a very simple quantum operation"""
import time

import qiskit.circuit as circuit
import qiskit.compiler as compiler

from tergite import Job, Tergite

if __name__ == "__main__":
    # the Tergite API URL e.g. "https://api.tergite.example"
    API_URL = "https://api.qal9000.se"
    # API token for connecting to tergite
    API_TOKEN = "<your Tergite API key >"
    # The name of the Quantum Computer to use from the available quantum computers
    BACKEND_NAME = "qiskit_pulse_2q"
    # the name of this service. For your own bookkeeping.
    SERVICE_NAME = "local"
    # the timeout in seconds for how long to keep checking for results
    POLL_TIMEOUT = 100

    # create the Qiskit circuit
    qc = circuit.QuantumCircuit(1)
    qc.x(0)
    qc.h(0)
    qc.measure_all()

    # create a provider
    # provider account creation can be skipped in case you already saved
    # your provider account to the `~/.qiskit/tergiterc` file.
    # See below how that is done.
    provider = Tergite.use_provider_account(
        service_name=SERVICE_NAME, url=API_URL, token=API_TOKEN
    )
    # to save this account to the `~/.qiskit/tergiterc` file, add the `save=True`
    # provider = Tergite.use_provider_account(account, save=True)

    # Get the tergite backend in case you skipped provider account creation
    # provider = Tergite.get_provider(service_name=SERVICE_NAME)
    backend = provider.get_backend(BACKEND_NAME)
    backend.set_options(shots=1024)

    # compile the circuit
    tc = compiler.transpile(qc, backend=backend)

    # run the circuit
    job: Job = backend.run(tc, meas_level=2, meas_return="single")
    job.wait_for_final_state(timeout=POLL_TIMEOUT)
    result = job.result()

    print(result.get_counts())



# This code is part of Tergite
#
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""tests for the running of qiskit circuits on the tergite backend"""
import json
import uuid
import warnings
from collections import Counter
from typing import List, Optional

import numpy as np
import pytest
from qiskit import QuantumCircuit, circuit, compiler, pulse
from qiskit.providers import JobStatus
from qiskit.result import Result
from qiskit.result.models import ExperimentResult, ExperimentResultData

from tergite import Job, OpenPulseBackend, Provider, Tergite

# cross compatibility with future qiskit version where deprecated packages are removed
from tergite.compat.qiskit.compiler.assembler import assemble
from tergite.services.api_client.dtos import DeviceCalibration, TergiteBackendConfig
from tergite.services.device_compiler.schedules import cz
from tests.utils.records import get_record
from tests.utils.requests import MockRequest, get_request_list

from ..utils.env import is_end_to_end
from .conftest import (
    API_TOKEN,
    API_URL,
    BACKENDS_LIST,
    GOOD_BACKENDS,
    INVALID_API_TOKENS,
    NUMBER_OF_SHOTS,
    QUANTUM_COMPUTER_URL,
    TEST_CALIBRATIONS_MAP,
    TEST_JOB_ID,
    TEST_JOB_RESULTS,
    TEST_QOBJ_ID,
    TEST_QOBJ_RESULTS_MAP,
    TWO_QUBIT_BACKENDS,
)

_INVALID_PARAMS = [
    (token, backend) for backend in GOOD_BACKENDS for token in INVALID_API_TOKENS
]


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_transpile_1q_gates(api, backend_name):
    """compiler.transpile(qc, backend=backend) returns backend-specific QuantumCircuits for 1-qubit ops"""
    backend = _get_backend(name=backend_name)
    calibrations = _get_calibrations(backend_name)
    qc = _get_test_1q_qiskit_circuit()

    # Transpile the circuit
    got = compiler.transpile(qc, backend=backend, initial_layout=qc.qubits)
    expected = _get_expected_1q_transpiled_circuit(
        backend=backend, calibrations=calibrations, circuit_name=got.name
    )

    got_qobj = backend.make_qobj(got)
    expected_qobj = backend.make_qobj(expected, qobj_id=got_qobj.qobj_id)

    assert (
        got_qobj == expected_qobj
    ), "Transpiled circuit does not match expected result."


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", TWO_QUBIT_BACKENDS)
def test_transpile_2q_gates(api, backend_name):
    """compiler.transpile(qc, backend=backend) returns backend-specific QuantumCircuits for 2-qubit gate ops"""
    backend = _get_backend(name=backend_name)
    calibrations = _get_calibrations(backend_name)
    qc = _get_test_2q_qiskit_circuit()
    expected = _get_expected_2q_transpiled_circuit(
        backend=backend, calibrations=calibrations, circuit_name=qc.name
    )

    # Transpile the circuit
    got = compiler.transpile(qc, backend=backend, initial_layout=qc.qubits)

    got_qobj = backend.make_qobj(got)
    expected_qobj = backend.make_qobj(expected, qobj_id=got_qobj.qobj_id)

    assert (
        got_qobj == expected_qobj
    ), "Transpiled circuit does not match expected result."


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_run_1q_gates(api, backend_name):
    """backend.run returns a registered job for 1-qubit gate operations"""
    backend = _get_backend(backend_name)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    backend.set_options(shots=NUMBER_OF_SHOTS)
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    qobj_id = str(uuid.uuid4())
    expected = _get_expected_job(
        backend=backend, transpiled_circuit=tc, meas_level=2, qobj_id=qobj_id
    )
    expected._calibration_date = calibration_date

    got = backend.run(tc, meas_level=2, qobj_id=qobj_id)
    requests_made = get_request_list(api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[:4]

    assert got == expected
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", TWO_QUBIT_BACKENDS)
def test_run_2q_gates(api, backend_name):
    """backend.run returns a registered job for 2-qubit gate operations"""
    backend = _get_backend(backend_name)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    backend.set_options(shots=NUMBER_OF_SHOTS)
    tc = _get_expected_2q_transpiled_circuit(backend=backend, calibrations=calibrations)
    qobj_id = str(uuid.uuid4())
    expected = _get_expected_job(
        backend=backend, transpiled_circuit=tc, meas_level=2, qobj_id=qobj_id
    )
    expected._calibration_date = calibration_date

    got = backend.run(tc, meas_level=2, qobj_id=qobj_id)
    requests_made = get_request_list(api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[:4]

    assert got == expected
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_run_bearer_auth(bearer_auth_api, backend_name):
    """backend.run returns a registered job for API behind bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    backend.set_options(shots=NUMBER_OF_SHOTS)
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    qobj_id = str(uuid.uuid4())
    expected = _get_expected_job(
        backend=backend, transpiled_circuit=tc, meas_level=2, qobj_id=qobj_id
    )
    expected._calibration_date = calibration_date

    got = backend.run(tc, meas_level=2, qobj_id=qobj_id)
    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[:4]

    assert got == expected
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("token, backend_name", _INVALID_PARAMS)
def test_run_invalid_bearer_auth(token, backend_name, bearer_auth_api):
    """backend.run with invalid bearer auth raises RuntimeError if backend is shielded with bearer auth"""
    backend = _get_backend(backend_name, token=token)
    backend.set_options(shots=NUMBER_OF_SHOTS)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    qobj_id = str(uuid.uuid4())

    with pytest.raises(
        RuntimeError,
        match=f"failed to get device calibrations for '{backend_name}': Unauthorized",
    ):
        _ = backend.run(tc, meas_level=2, qobj_id=qobj_id)

    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:2]

    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_result(api, backend_name):
    """job.result() returns a successful job's results"""
    backend = _get_backend(backend_name)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    expected = _get_expected_job_result(backend=backend, job=job)

    got = job.result()
    requests_made = get_request_list(api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert got.to_dict() == expected.to_dict()
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_result_bearer_auth(bearer_auth_api, backend_name):
    """job.result() returns a successful job's results for API behind bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    expected = _get_expected_job_result(backend=backend, job=job)
    got = job.result()
    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert got.to_dict() == expected.to_dict()
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("token, backend_name", _INVALID_PARAMS)
def test_job_result_invalid_bearer_auth(token, backend_name, bearer_auth_api):
    """job.result() with invalid bearer auth raises RuntimeError if backend is shielded with bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    # change the token to the invalid one
    backend.provider.account.token = token

    with pytest.raises(RuntimeError, match=f"error retrieving job data: Unauthorized"):
        _ = job.result()

    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_status(api, backend_name):
    """job.status() returns a successful job's status"""
    backend = _get_backend(backend_name)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    got = job.status()
    requests_made = get_request_list(api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert got == JobStatus.DONE
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_status_bearer_auth(bearer_auth_api, backend_name):
    """job.status() returns a successful job's status for API behind bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    got = job.status()
    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert got == JobStatus.DONE
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("token, backend_name", _INVALID_PARAMS)
def test_job_status_invalid_bearer_auth(token, backend_name, bearer_auth_api):
    """job.status() with invalid bearer auth raises RuntimeError if backend is shielded with bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    # change the token to the invalid one
    backend.provider.account.token = token

    with pytest.raises(RuntimeError, match=f"error retrieving job data: Unauthorized"):
        _ = job.status()

    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_cancel(api, backend_name):
    """job.cancel() cancels the running job"""
    backend = _get_backend(backend_name)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)
    job.cancel()

    mock_cancel_request = _get_mock_cancel_request(job)
    requests_made = get_request_list(api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:4] + [mock_cancel_request]

    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_cancel_bearer_auth(bearer_auth_api, backend_name):
    """job.cancel() calls the cancel endpoint for API behind bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)
    job.cancel()

    mock_cancel_request = _get_mock_cancel_request(job)
    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:4] + [mock_cancel_request]

    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("token, backend_name", _INVALID_PARAMS)
def test_job_cancel_invalid_bearer_auth(token, backend_name, bearer_auth_api):
    """job.cancel() with invalid bearer auth raises RuntimeError if backend is shielded with bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)
    job_id = job.job_id()

    # change the token to the invalid one
    backend.provider.account.token = token

    with pytest.raises(
        RuntimeError, match=f"Failed to cancel job '{job_id}': Unauthorized"
    ):
        _ = job.cancel()

    mock_cancel_request = _get_mock_cancel_request(job)
    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:4] + [mock_cancel_request]

    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_download_url(api, backend_name):
    """job.download_url returns a successful job's download_url"""
    backend = _get_backend(backend_name)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    got = job.download_url
    requests_made = get_request_list(api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert got == TEST_JOB_RESULTS["download_url"]
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_download_url_bearer_auth(bearer_auth_api, backend_name):
    """job.download_url returns a successful job's download_url for API behind bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    got = job.download_url
    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert got == TEST_JOB_RESULTS["download_url"]
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("token, backend_name", _INVALID_PARAMS)
def test_job_download_url_invalid_bearer_auth(token, backend_name, bearer_auth_api):
    """job.download_url with invalid bearer auth raises RuntimeError if backend is shielded with bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    # change the token to the invalid one
    backend.provider.account.token = token

    with pytest.raises(RuntimeError, match=f"error retrieving job data: Unauthorized"):
        _ = job.download_url

    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_logfile(api, backend_name, tmp_results_file):
    """job.logfile downloads a job's data to tmp"""
    backend = _get_backend(backend_name)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    assert job.logfile == tmp_results_file
    requests_made = get_request_list(api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:6]

    with open(tmp_results_file, "rb") as file:
        got = json.load(file)

    assert got == TEST_JOB_RESULTS
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_job_logfile_bearer_auth(bearer_auth_api, backend_name, tmp_results_file):
    """job.logfile downloads a successful job's results for API behind bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    assert job.logfile == tmp_results_file
    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:6]

    with open(tmp_results_file, "rb") as file:
        got = json.load(file)

    assert got == TEST_JOB_RESULTS
    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("token, backend_name", _INVALID_PARAMS)
def test_job_logfile_invalid_bearer_auth(token, backend_name, bearer_auth_api):
    """job.logfile with invalid bearer auth raises RuntimeError if backend is shielded with bearer auth"""
    backend = _get_backend(backend_name, token=API_TOKEN)
    calibrations = _get_calibrations(backend_name)
    calibration_date = calibrations.last_calibrated
    tc = _get_expected_1q_transpiled_circuit(backend=backend, calibrations=calibrations)
    job = backend.run(tc, meas_level=2)

    # change the token to the invalid one
    backend.provider.account.token = token

    with pytest.raises(RuntimeError, match=f"error retrieving job data: Unauthorized"):
        _ = job.logfile

    requests_made = get_request_list(bearer_auth_api)
    expected_requests = _get_all_mock_requests(
        backend_name, calibration_date=calibration_date
    )[1:5]

    assert requests_made == expected_requests


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_provider_job(api_with_logfile, backend_name):
    """Test that Provider.job() returns the correct Job object."""
    provider = Tergite.use_provider_account(service_name="test", url=API_URL)

    # create a job the usual way
    backend = _get_backend(backend_name, provider=provider)
    backend.set_options(shots=NUMBER_OF_SHOTS)
    calibrations = _get_calibrations(backend_name)
    qobj_id = f"{TEST_QOBJ_ID}-{backend_name}"
    circuit_name = TEST_QOBJ_RESULTS_MAP[backend_name.lower()]["experiments"][0][
        "header"
    ]["name"]

    tc = _get_expected_1q_transpiled_circuit(
        backend=backend, calibrations=calibrations, circuit_name=circuit_name
    )
    expected = backend.run(tc, meas_level=2, qobj_id=qobj_id)

    assert expected.logfile is not None
    # overwrite some properties that are expected to change when getting job by job id
    expected.upload_url = expected.metadata["upload_url"] = None
    job_id = expected.job_id()

    # retrieve job from provider
    got = provider.job(job_id)

    assert got == expected, "The retrieved job does not match the original job."


def _get_expected_job_result(backend: OpenPulseBackend, job: Job) -> Result:
    """Returns the expected job result"""
    results = [
        ExperimentResult(
            header=job.payload.experiments[index].header,
            shots=job.metadata["shots"],
            success=True,
            data=ExperimentResultData(
                counts=dict(Counter(result)),
                memory=result,
            ),
        )
        for index, result in enumerate(TEST_JOB_RESULTS["result"]["memory"])
    ]

    return Result(
        backend_name=backend.name,
        backend_version=backend.backend_version,
        qobj_id=job.metadata["qobj_id"],
        job_id=job.job_id(),
        success=True,
        results=results,
    )


def _get_expected_job(
    backend: OpenPulseBackend,
    transpiled_circuit: QuantumCircuit,
    qobj_id: str,
    **options,
) -> Job:
    """Returns the expected job after being initialized"""
    schedule = compiler.schedule(transpiled_circuit, backend=backend)
    with warnings.catch_warnings():
        # The class QobjExperimentHeader is deprecated
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="qiskit")
        qobj = assemble(
            experiments=[schedule],
            backend=backend,
            shots=NUMBER_OF_SHOTS,
            qubit_lo_freq=backend.qubit_lo_freq,
            meas_lo_freq=backend.meas_lo_freq,
            qobj_id=qobj_id,
            **options,
        )

    job = Job(
        backend=backend,
        job_id=TEST_JOB_ID,
        payload=qobj,
        upload_url=f"{QUANTUM_COMPUTER_URL}/jobs",
    )

    job.metadata["shots"] = NUMBER_OF_SHOTS
    job.metadata["qobj_id"] = qobj_id
    job.metadata["num_experiments"] = 1

    return job


def _get_test_1q_qiskit_circuit():
    """Returns a qiskit quantum circuit for testing two qubit gates"""
    qc = circuit.QuantumCircuit(1)
    qc.h(0)
    qc.measure_all()
    return qc


def _get_test_2q_qiskit_circuit():
    """Returns a qiskit quantum circuit for testing for testing one qubit gate"""
    qc = circuit.QuantumCircuit(2)
    qc.h(0)
    qc.h(1)
    qc.cx(0, 1)
    qc.measure_all()
    return qc


def _get_expected_1q_transpiled_circuit(
    backend: OpenPulseBackend,
    calibrations: DeviceCalibration,
    circuit_name: Optional[str] = None,
) -> circuit.QuantumCircuit:
    """Returns a quantum circuit for 1-qubit gates specific to the TEST_BACKEND

    Args:
        backend: the backend for which the circuit is transpiled
        calibrations: the device parameters for the given backend
        circuit_name: the name of the expected circuit

    Returns:
        The circuit.QuantumCircuit that corresponds to the 1-qubit gate example
    """
    phase = np.pi / 2
    qc = circuit.QuantumCircuit(1, global_phase=phase, name=circuit_name)
    qc.rz(phase, 0)
    qc.rx(phase, 0)
    qc.rz(phase, 0)
    qc.measure_all()

    # initialize calibrations
    qubit_0 = calibrations.qubits[0]

    rz_block = pulse.ScheduleBlock(
        name="RZ(λ, (0,))",
        alignment_context=pulse.transforms.AlignLeft(),
    )
    rz_block.append(
        pulse.ShiftPhase(round(phase, 10), pulse.DriveChannel(0), name="RZ q0")
    )

    rx_block = pulse.ScheduleBlock(
        name="RX(θ, (0,))",
        alignment_context=pulse.transforms.AlignLeft(),
    )
    rx_block.append(pulse.SetFrequency(qubit_0.frequency.value, pulse.DriveChannel(0)))
    rx_block.append(
        pulse.Play(
            # amp represents the magnitude of the complex amplitude and can't be complex
            pulse.Gaussian(
                duration=round(qubit_0.pi_pulse_duration.value / backend.dt),
                amp=round(phase / np.pi * qubit_0.pi_pulse_amplitude.value, 10),
                sigma=round(qubit_0.pulse_sigma.value / backend.dt),
                name="RX q0",
            ),
            pulse.DriveChannel(0),
            name="RX q0",
        )
    )

    qc._calibrations = {
        "rz": {((0,), (phase,)): rz_block},
        "rx": {((0,), (phase,)): rx_block},
    }

    return qc


def _get_expected_2q_transpiled_circuit(
    backend: OpenPulseBackend,
    calibrations: DeviceCalibration,
    circuit_name: Optional[str] = None,
):
    """Returns a quantum circuit for 2-qubit gates specific to the TEST_BACKEND

    Args:
        backend: the backend for which the circuit is transpiled
        calibrations: the device parameters for the given backend
        circuit_name: the name of the expected circuit

    Returns:
        The circuit.QuantumCircuit that corresponds to the 2-qubit gate example
    """
    phase = np.pi / 2
    qc = circuit.QuantumCircuit(2, global_phase=np.pi, name=circuit_name)
    qc.rz(phase, 0)
    qc.rx(phase, 0)
    qc.rz(phase, 0)

    qc.cz(0, 1)

    qc.rz(phase, 1)
    qc.rx(phase, 1)
    qc.rz(phase, 1)

    qc.measure_all()

    ##
    qubit_0 = calibrations.qubits[0]
    qubit_1 = calibrations.qubits[1]

    # initialize calibrations
    # rz_block_0
    rz_block_0 = pulse.ScheduleBlock(
        name="RZ(λ, (0,))",
        alignment_context=pulse.transforms.AlignLeft(),
    )
    rz_block_0.append(
        pulse.ShiftPhase(round(phase, 10), pulse.DriveChannel(0), name="RZ q0")
    )

    # rz_block_1
    rz_block_1 = pulse.ScheduleBlock(
        name="RZ(λ, (1,))",
        alignment_context=pulse.transforms.AlignLeft(),
    )
    rz_block_1.append(
        pulse.ShiftPhase(round(phase, 10), pulse.DriveChannel(1), name="RZ q1")
    )

    # rx_block_0
    rx_block_0 = pulse.ScheduleBlock(
        name="RX(θ, (0,))",
        alignment_context=pulse.transforms.AlignLeft(),
    )
    rx_block_0.append(
        pulse.SetFrequency(qubit_0.frequency.value, pulse.DriveChannel(0))
    )
    rx_block_0.append(
        pulse.Play(
            # amp represents the magnitude of the complex amplitude and can't be complex
            pulse.Gaussian(
                duration=round(qubit_0.pi_pulse_duration.value / backend.dt),
                amp=round(phase / np.pi * qubit_0.pi_pulse_amplitude.value, 10),
                sigma=round(qubit_0.pulse_sigma.value / backend.dt),
                name="RX q0",
            ),
            pulse.DriveChannel(0),
            name="RX q0",
        )
    )

    # rx_block_1
    rx_block_1 = pulse.ScheduleBlock(
        name="RX(θ, (1,))",
        alignment_context=pulse.transforms.AlignLeft(),
    )
    rx_block_1.append(
        pulse.SetFrequency(qubit_1.frequency.value, pulse.DriveChannel(1))
    )
    rx_block_1.append(
        pulse.Play(
            # amp represents the magnitude of the complex amplitude and can't be complex
            pulse.Gaussian(
                duration=round(qubit_1.pi_pulse_duration.value / backend.dt),
                amp=round(phase / np.pi * qubit_1.pi_pulse_amplitude.value, 10),
                sigma=round(qubit_1.pulse_sigma.value / backend.dt),
                name="RX q1",
            ),
            pulse.DriveChannel(1),
            name="RX q1",
        )
    )

    # cz_block
    cz_block = cz(
        backend=backend,
        control_qubit_idxs=(0,),
        target_qubit_idxs=(1,),
        device_properties=calibrations,
    )

    qc._calibrations = {
        "cz": {((0, 1), ()): cz_block},
        "rz": {((0,), (phase,)): rz_block_0, ((1,), (phase,)): rz_block_1},
        "rx": {((0,), (phase,)): rx_block_0, ((1,), (phase,)): rx_block_1},
    }

    return qc


def _get_backend(name: str, token: str = None, provider: Optional[Provider] = None):
    """Retrieves the right backend

    Args:
        name: the name of the backend
        token: the API token to use to access the API; defaults to None
        provider: an optional provider to use; defaults to None

    Returns:
        An OpenPulseBackend corresponding to the given name
    """
    if provider is None:
        provider = Tergite.use_provider_account(
            service_name="test", url=API_URL, token=token
        )

    expected_json = get_record(BACKENDS_LIST, _filter={"name": name})
    return OpenPulseBackend(
        data=TergiteBackendConfig(**expected_json), provider=provider, base_url=API_URL
    )


def _get_calibrations(backend_name: str) -> DeviceCalibration:
    """Retrieves the device calibrations for the given device

    Args:
        backend_name: the name of the device

    Returns:
        the DeviceCalibration of the given device
    """
    data = TEST_CALIBRATIONS_MAP[backend_name]
    return DeviceCalibration(**data)


def _get_all_mock_requests(
    backend_name: str, calibration_date: Optional[str] = None
) -> List[MockRequest]:
    """Generates all the possible mock requests for a given backend

    Args:
        backend_name: the name of the backend

    Returns:
        The list of all MockRequests for the given backend name
    """
    return [
        MockRequest(
            url=f"https://api.tergite.example/calibrations/{backend_name}",
            method="GET",
        ),
        MockRequest(
            url=f"https://api.tergite.example/calibrations/{backend_name}",
            method="GET",
        ),
        MockRequest(
            url=f"https://api.tergite.example/jobs/",
            method="POST",
            json={"device": backend_name, "calibration_date": calibration_date},
            has_text=True,
        ),
        MockRequest(
            url="http://loke.tergite.example/jobs", method="POST", has_text=True
        ),
        MockRequest(
            url="https://api.tergite.example/jobs/test_job_id",
            method="GET",
            has_text=False,
        ),
        MockRequest(url="http://loke.tergite.example/test_file.hdf5", method="GET"),
    ]


def _get_mock_cancel_request(job: Job) -> MockRequest:
    """Gets the mock cancel request for the given job

    Args:
        job: the job to cancel

    Returns:
        the MockRequest
    """
    return MockRequest(
        url=f"http://loke.tergite.example/jobs/{job.job_id()}/cancel",
        method="POST",
        json={},
        has_text=True,
    )



from qiskit import *

from tergite.qiskit.providers import Tergite

provider = Tergite.get_provider()
pingu_backend = provider.get_backend("pingu")

quantum_register = QuantumRegister(3)  # 3 qubits
quantum_circuit = QuantumCircuit(quantum_register)

quantum_circuit.x(quantum_register[1])
quantum_circuit.h(quantum_register[2])

print(quantum_circuit)



import matplotlib.pyplot as plt
import numpy as np

x = np.linspace(0, 6 * np.pi, 100)
y = np.sin(x)

# You probably won't need this if you're embedding things in a tkinter plot...
plt.ion()

fig = plt.figure()
ax = fig.add_subplot(111)
(line1,) = ax.plot(x, y, "r-")  # Returns a tuple of line objects, thus the comma

for phase in np.linspace(0, 10 * np.pi, 500):
    line1.set_ydata(np.sin(x + phase))
    fig.canvas.draw()
    fig.canvas.flush_events()



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


# this file is meant to be run interactively for objects inspection
# python -i tests/job_on_simulator.py

from qiskit import *

backend = BasicAer.get_backend("qasm_simulator")

circ = QuantumCircuit(3, 3)

circ.x(2)
circ.h(1)
circ.measure([0, 1, 2], [0, 1, 2])

job = execute(circ, backend)
counts = job.result().get_counts()

print("Counts:", counts)
# plot_histogram(counts).show()



from datetime import datetime

import torch
import torch.nn as nn
from data_handling import IQDataGenerator
from learning import Net
from torch import optim
from torch.utils.data import DataLoader
from tqdm import tqdm
from utils import Utils

RUN_ID = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
N_QUBITS = 2
TRAIN_EPOCHS = 10

N_TRAIN_SAMPLES = 100
N_EVAL_SAMPLES = 25

BACKEND = Utils.get_backend("AerSimulator")

# generate data
train_loader = DataLoader(
    IQDataGenerator.get_2d_binary_class(n_samples=N_TRAIN_SAMPLES),
    batch_size=1,
    shuffle=True,
)
eval_loader = DataLoader(
    IQDataGenerator.get_2d_binary_class(n_samples=N_EVAL_SAMPLES),
    batch_size=1,
    shuffle=True,
)

# init model
model = Net(RUN_ID, n_qubits=N_QUBITS, backend=BACKEND)
optimizer = optim.Adam(model.parameters(), lr=0.001)
loss_func = nn.NLLLoss()

loss_list = []

# train
model.train()
for epoch in range(TRAIN_EPOCHS):
    total_loss = []
    for batch_idx, (data, target) in tqdm(
        enumerate(train_loader), desc=f"{epoch = }", unit="samples"
    ):
        optimizer.zero_grad()
        output = model(data)
        loss = loss_func(output, target)
        loss.backward()
        optimizer.step()

        total_loss.append(loss.item())

    loss_list.append(sum(total_loss) / len(total_loss))

    print(
        "\nTraining [{:.0f}%]\tLoss: {:.4f}".format(
            100.0 * (epoch + 1) / TRAIN_EPOCHS, loss_list[-1]
        )
    )

# save
torch.save(model, f"{RUN_ID}_example.model")

# load
loaded_model = torch.load(f"{RUN_ID}_example.model")

# eval
loaded_model.eval()
with torch.no_grad():
    correct = 0
    for batch_idx, (data, target) in enumerate(eval_loader):
        output = loaded_model(data)

        pred = output.argmax(dim=1, keepdim=True)
        correct += pred.eq(target.view_as(pred)).sum().item()

        loss = loss_func(output, target)
        total_loss.append(loss.item())

    print(
        "Performance on test data:\n\tLoss: {:.4f}\n\tAccuracy: {:.1f}%".format(
            sum(total_loss) / len(total_loss), correct / len(eval_loader) * 100
        )
    )

# states = []
# thetas = Utils.load_thetas(RUN_ID)
# print(len(thetas))
# for theta in thetas:
#     states.append((cos(theta / 2) * basis(2, 0) + sin(theta / 2) * basis(2, 1)).unit())
#
# animate_bloch(states, duration=0.1, save_all=False, filename=f'{RUN_ID}_blochsphere')



#!/usr/bin/env python3.9
# coding: utf-8
from datetime import datetime
from os import listdir, makedirs
from pathlib import Path
from shutil import move
from uuid import uuid4 as uuid

import matplotlib.pyplot as plt
import numpy as np
import requests
from tqdm.auto import tqdm

from tergite.qiskit.providers import Tergite

folder = Path("demo_qmc_frames").resolve()


def save_old_frames():
    global folder
    saved_folder = folder / "saved_animations"
    makedirs(saved_folder, exist_ok=True)

    old_frames = [f for f in listdir(folder) if f.endswith(".jpg")]

    if len(old_frames):
        now_time = datetime.now()
        mstr = now_time.strftime("%Y%m%d%H%M%S")
        new_dir = saved_folder / mstr
        makedirs(new_dir, exist_ok=True)
        for f in old_frames:
            move(folder / f, new_dir / f)


save_old_frames()


# -------------------------- HELPERS FOR GENERATING FIGURE ----------------------------------
def f(x):
    return np.sqrt(1 - x**2)


# hitmiss estimator, given input points
def qmc_estimate(fn: callable, U: np.ndarray) -> float:
    estimate = np.mean(
        U[:, 1] <= fn(U[:, 0])
    )  # <- This is the estimator for integral value
    return estimate


def plot_labels_etc(ax: object):
    text_settings = dict(fontsize=23)
    ax.set_title("Quantum Monte Carlo", **text_settings)
    ax.plot(np.linspace(0, 1, 1000), f(np.linspace(0, 1, 1000)), lw=3.5, zorder=99)
    ax.set_ylabel("f(x)", **text_settings)
    ax.set_xlabel("x", **text_settings)
    ax.grid(zorder=-99, alpha=0.5)
    ax.tick_params(axis="both", which="minor", labelsize=text_settings["fontsize"])
    ax.tick_params(axis="both", which="major", labelsize=text_settings["fontsize"])


# plots a textbox
def plot_estimate(ax: object, estimate: float):
    textstr = "$\hat\pi = {:10.5f}$".format(estimate)
    props = dict(boxstyle="round", facecolor="wheat", alpha=1)
    # place a text box in upper left in axes coords
    ax.text(
        0.15,
        0.15,
        textstr,
        transform=ax.transAxes,
        verticalalignment="top",
        bbox=props,
        fontsize=25,
    )


# plots points with formatting
def plot_samples(ax: object, samples: np.array, marker: str = "o"):
    inside = np.sqrt(samples[:, 0] ** 2 + samples[:, 1] ** 2) < 1
    ax.scatter(samples[inside, 0], samples[inside, 1], color="C1", marker=marker)
    ax.scatter(samples[~inside, 0], samples[~inside, 1], color="C2", marker=marker)


# -------------------------- BACKEND SETTINGS ----------------------------------
vm_url = "http://qtl-axean.mc2.chalmers.se"
mss_url = vm_url + ":8002"
bcc_url = vm_url + ":8000"
chalmers = Tergite.get_provider()
backend = chalmers.get_backend("Nov7")
print(f"Connected to backend: {backend.name}")

# -------------------------- COMPUTING THE FRAMES ----------------------------------

points = np.asarray([[], []]).T
job_ids = list(str(uuid()) for _ in range(15))


def compute_new_frame(j: int):
    global points
    global job_ids
    job_id = job_ids[j]

    fig, ax = plt.subplots(figsize=(7.6, 7.6))

    # queue job on backend
    bcc_response = requests.get(bcc_url + "/rng/" + job_id)
    assert bcc_response.ok

    # wait for job to complete, blocking REST API call chain
    mss_response = requests.get(mss_url + "/rng/" + job_id)
    assert mss_response.ok
    data = mss_response.json()

    # split obtained numbers into two disjoint subsets
    N = data["N"] // 2
    X = np.asarray(data["numbers"][:N]).astype(float)  # first N are X coordinate
    Y = np.asarray(data["numbers"][N : (N + N)]).astype(
        float
    )  # second N are Y coordinate

    # these numbers are random 32-bit integers in the range [-2**31, 2**31 - 1]
    # so we want to convert them to floats in the range [0,1]
    X += 2**31
    Y += 2**31

    X *= 2**-32
    Y *= 2**-32

    # concatenate points obtained
    U = np.zeros((N, 2))
    U[:, 0] = X
    U[:, 1] = Y
    points = np.concatenate((points, U))
    est = qmc_estimate(f, points)

    # create figure and save
    plot_labels_etc(ax)
    plot_samples(ax, points)
    plot_estimate(ax, est * 4)

    fig.savefig(folder / f"frame{j}.jpg", dpi=300, bbox_inches="tight")


for j in tqdm(range(len(job_ids)), desc="Quantum Monte Carlo"):
    compute_new_frame(j)



# This code is part of Tergite
#
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import json
from os import path
from typing import Any, Dict, List, Union

_TESTS_FOLDER = path.dirname(path.dirname(path.abspath(__file__)))
_FIXTURES_PATH = path.join(_TESTS_FOLDER, "fixtures")


def load_json_fixture(file_name: str) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    """Loads the given fixture from the fixtures directory

    Args:
        file_name: the name of the file that contains the fixture

    Returns:
        the list of dicts or the dict got from the json fixture file
    """
    fixture_path = _get_fixture_path(file_name)
    with open(fixture_path, "rb") as file:
        return json.load(file)


def _get_fixture_path(*paths: str) -> str:
    """Gets the path to the fixture

    Args:
        paths: sections of paths to the given fixture e.g. fixtures/api/rest/many_rngs.json
            would be "api", "rest", "many_rngs.json"

    Returns:
        the path to the given fixture
    """
    return path.join(_FIXTURES_PATH, *paths)



# This code is part of Tergite
#
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Utilities for handling records in tests"""
from typing import Any, Dict, List


def get_record(data: List[Dict[str, Any]], _filter: Dict[str, Any]) -> Dict[str, Any]:
    """Gets the first record in data which matches the given filter

    Args:
        data: list of records to get the record from
        _filter: partial dict that the given record should contain

    Returns:
        the first record that matches the given filter

    Raises:
        KeyError: {key}
    """
    try:
        return next(
            filter(lambda x: all([x[k] == v for k, v in _filter.items()]), data)
        )
    except StopIteration:
        raise KeyError(f"no match found for filter {_filter}")


def get_many_records(
    data: List[Dict[str, Any]], _filter: Dict[str, Any]
) -> List[Dict[str, Any]]:
    """Gets all records in data that match the given filter

    Args:
        data: list of records to get the record from
        _filter: partial dict that the given record should contain

    Returns:
        the records that match the given filter

    Raises:
        KeyError: {key}
    """
    try:
        return list(
            filter(lambda x: all([x[k] == v for k, v in _filter.items()]), data)
        )
    except IndexError:
        raise KeyError(f"no match found for filter {_filter}")



import numpy as np
import torch
from torch.utils.data import Dataset


class IQDataset(Dataset):
    def __init__(self, data, targets):
        self.data = data
        self.targets = targets

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index], self.targets[index]


class IQDataGenerator:
    @staticmethod
    def get_2d_binary_class(
        n_samples=100,
        mean_zeros=np.array([1, 1]),
        mean_ones=np.array([-1, -2]),
        cov_zeros=np.array([[0.3, 0.6], [0.8, -0.4]]),
        cov_ones=np.array([[-0.4, 0.4], [0.7, 0.6]]),
    ):
        zeros = np.random.multivariate_normal(mean_zeros, cov_zeros, n_samples)
        ones = np.random.multivariate_normal(mean_ones, cov_ones, n_samples)

        datapoints = np.append(zeros, ones).reshape((n_samples * 2, 2))
        targets = np.append(
            [0 for _ in range(n_samples)], [1 for _ in range(n_samples)]
        )

        datapoints_tensor = torch.from_numpy(datapoints).float()
        targets_tensor = torch.from_numpy(targets).long()

        return IQDataset(datapoints_tensor, targets_tensor)



#!/usr/bin/env python3.9
# coding: utf-8

from os import system
from pathlib import Path

import matplotlib.pyplot as plt

plt.ion()  # enable interactive mode

folder = Path(input("Frame folder? ")).resolve()

fig, ax = plt.subplots(figsize=(4.6, 4.6))

# show first frame, draw interactively on this image object
im = ax.imshow(plt.imread(Path(input("First frame? ")).resolve()))
ax.axis("off")


def draw_frame(fig, ax, frame_path: Path):
    global im

    image_j = plt.imread(frame_path)

    # show next frame
    im.set_array(image_j)

    fig.canvas.draw()
    fig.canvas.flush_events()


frames = iter(folder / f"frame{j}.jpg" for j in range(1000))
frame = next(frames)

while True:
    wait_for_me = input(f"Draw {frame.name}? ")
    if wait_for_me:
        if str.lower(wait_for_me) == "exit":
            # end program on 'exit', stop drawing
            break
        else:
            # wait again on not None, also on not 'exit'
            continue
    try:
        draw_frame(fig, ax, frame)
        drew = True

    # try to draw. if you can't then just ask for input again
    except Exception as err:
        drew = False
    finally:
        system("cls || clear")

    if drew:
        frame = next(frames)
    else:
        frame = frame


plt.show()



import time
from uuid import uuid4 as uuid

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import requests
from qiskit.providers.jobstatus import JobStatus
from qiskit.providers.tergite import Tergite

# Task: Estimate Pi numerically
# Area of circle = pi*r**2
#
# If we compute the area of a quarter of a circle and multiply it by 4, we should get Pi = 3.14159...
# The size of the circle doesn't matter, so let's take the top right quadrant of the unit circle.

# ---------------------- settings
vm_url = "http://qtl-axean.mc2.chalmers.se"
mss_url = vm_url + ":8002"
bcc_url = vm_url + ":8000"
chalmers = Tergite.get_provider()
backend = chalmers.get_backend("Nov7")
print(f"Connected to backend: {backend.name}")

text_settings = {"fontsize": 20}
plt.ion()  # enable interactive mode
fig, axs = plt.subplots(1, 3, figsize=(17, 6.5), sharex=True, sharey=True)
# ---------------------- settings


def f(x):
    return np.sqrt(1 - x**2)


def mc_hitmiss(fn: callable, *, M: int) -> tuple:
    np.random.seed(0)
    U = np.random.rand(M, 2)
    estimate = np.mean(
        U[:, 1] <= fn(U[:, 0])
    )  # <- This is the estimator for integral value
    return estimate, U


def plot_estimate(ax: object, estimate: float):
    textstr = "$\hat\pi = {:10.5f}$".format(estimate)
    props = dict(boxstyle="round", facecolor="wheat", alpha=0.95)
    # place a text box in upper left in axes coords
    ax.text(
        0.15,
        0.15,
        textstr,
        transform=ax.transAxes,
        verticalalignment="top",
        bbox=props,
        fontsize=25,
    )


def plot_samples(ax: object, samples: np.array, marker: str = "o"):
    inside = np.sqrt(samples[:, 0] ** 2 + samples[:, 1] ** 2) < 1
    ax.scatter(samples[inside, 0], samples[inside, 1], color="C1", marker=marker)
    ax.scatter(samples[~inside, 0], samples[~inside, 1], color="C2", marker=marker)


def api_error(url, response):
    return f"Failed to communicate with {url}, response: {response}"


def wait_for_job(job_id: str, timeout: int = 20) -> object:
    mss_response = requests.get(mss_url + "/rng/" + job_id)
    return mss_response


def qmc_estimate(fn: callable, U: np.ndarray) -> float:
    estimate = np.mean(
        U[:, 1] <= fn(U[:, 0])
    )  # <- This is the estimator for integral value
    return estimate


def qmc_hitmiss_parts(fn: callable, num_jobs: int = 3) -> tuple:
    job_ids = list(str(uuid()) for _ in range(num_jobs))

    estimates = list()  # should be monotonic increasing
    points = np.asarray([[], []]).T
    points_by_job = dict()

    for job_id in job_ids:
        # queue job on backend directly
        bcc_response = requests.get(bcc_url + "/rng/" + job_id)
        if not bcc_response.ok:
            raise RuntimeError(api_error(bcc_url, bcc_response))

        # wait for job to complete
        mss_response = wait_for_job(job_id)
        data = mss_response.json()

        # split obtained numbers into two disjoint subsets
        N = data["N"] // 2
        X = np.asarray(data["numbers"][:N]).astype(float)  # first M are X coordinate
        Y = np.asarray(data["numbers"][N : (N + N)]).astype(
            float
        )  # second M are Y coordinate

        # these numbers are random 32-bit integers in the range [-2**31, 2**31 - 1]
        # so we want to convert them to floats in the range [0,1]
        X += 2**31
        Y += 2**31
        X *= 2**-32
        Y *= 2**-32

        # concatenate points obtained
        U = np.zeros((N, 2))
        U[:, 0] = X
        U[:, 1] = Y
        points = np.concatenate((points, U))
        points_by_job[job_id] = U
        est = qmc_estimate(fn, points)

        plot_samples(axs[2], U)
        plot_estimate(axs[2], est * 4)
        fig.canvas.draw()
        fig.canvas.flush_events()

        # compute estimator on all the points, up until now
        estimates.append(est)

    return estimates, points_by_job, job_ids


xdef = np.linspace(0, 1)

# ---------- Riemann estimate
riemann_est = sum(f(xdef) * np.diff(xdef)[0])
riemann_est_pi = riemann_est * 4

# ---------- Monte Carlo estimate, from pseudo random numbers
mc_est, mc_samples = mc_hitmiss(f, M=400)
mc_est_pi = mc_est * 4

# ---------- Plotting non quantum stuff
# plot riemann rectangles
riemann_rects = [
    mpl.patches.Rectangle(
        xy=(x, 0),
        width=np.diff(xdef)[0],
        height=f(x),
        ec="black",
        fc="none",
        alpha=0.75,
    )
    for x in xdef
]
for r in riemann_rects:
    axs[0].add_patch(r)
plot_estimate(axs[0], riemann_est_pi)

axs[0].set_title("Riemann", **text_settings)
axs[1].set_title("Monte Carlo", **text_settings)
axs[2].set_title("Quantum Monte Carlo", **text_settings)

for ax in axs:
    ax.plot(np.linspace(0, 1, 1000), f(np.linspace(0, 1, 1000)), lw=3.5, zorder=99)
    ax.set_ylabel("f(x)", **text_settings)
    ax.set_xlabel("x", **text_settings)
    ax.grid(zorder=-99, alpha=0.5)
    ax.tick_params(axis="both", which="minor", labelsize=text_settings["fontsize"])
    ax.tick_params(axis="both", which="major", labelsize=text_settings["fontsize"])
fig.suptitle("$f(x) = (1-x^2)^{1/2}$", **text_settings)

plot_samples(axs[1], mc_samples)
plot_estimate(axs[1], mc_est_pi)

fig.canvas.draw()
fig.canvas.flush_events()

# ---------- Plotting quantum stuff
qmc_hitmiss_parts(f, num_jobs=15)



from qiskit import *
from qiskit.visualization import plot_histogram

from tergite.qiskit.providers import Tergite

IBMQ.load_account()

tergite_provider = Tergite.get_provider()
ibmq_provider = IBMQ.get_provider()
sim_provider = BasicAer

pingu_backend = tergite_provider.get_backend("pingu")
ibmq_backend = ibmq_provider.get_backend("ibmq_burlington")
sim_backend = sim_provider.get_backend("qasm_simulator")

quantum_circuit = QuantumCircuit(3, 3)

quantum_circuit.h(1)
quantum_circuit.x(2)
quantum_circuit.t(2)

quantum_circuit.measure([0, 1, 2], [0, 1, 2])
print(quantum_circuit)

job_pingu = execute(quantum_circuit, pingu_backend)
job_sim = execute(quantum_circuit, sim_backend)
job_ibmq = execute(quantum_circuit, ibmq_backend)

print("Waiting for IBMQ results", flush=True)
job_ibmq.wait_for_final_state()

c_sim = job_sim.result().get_counts()
c_pingu = job_pingu.result().get_counts()
c_ibmq = job_ibmq.result().get_counts()

plot_histogram(c_sim).show()
plot_histogram(c_pingu).show()
plot_histogram(c_ibmq).show()



# This code is part of Tergite
#
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Utilities for handling mock requests in tests"""
import dataclasses
import json
from typing import Any, Dict, List, Optional


@dataclasses.dataclass
class MockRequest:
    url: str
    method: str
    json: Optional[Dict[str, Any]] = None
    has_text: bool = False

    @classmethod
    def load(cls, req):
        json_data = None
        try:
            json_data = req.json()
        except (AttributeError, TypeError, json.JSONDecodeError):
            pass

        return cls(
            url=req.url,
            method=req.method,
            json=json_data,
            has_text=(req.text is not None),
        )


def get_request_list(requests_mocker) -> List[MockRequest]:
    """Retrieves the list of requests from the requests_mocker.

    Args:
        requests_mocker: the Mock object from requests_mock

    Returns:
        the list of MockRequest made
    """

    return [MockRequest.load(v) for v in requests_mocker.request_history]



# This code is part of Tergite
#
# (C) Copyright Axel Andersson 2022
# (C) Copyright Chalmers Next Labs 2025
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Utility function for handling hte Qobj for a job"""
from itertools import groupby
from typing import Any, Dict, List, Tuple, Union


def compress_qobj_dict(qobj_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Creates a compressed dictionary representation of the qobj dict

    In order to reduce the bandwidth taken up be the qobject
    dict, we do a few things with the data which will be reversed
    at the backend

    Note that this compression is in-place

    Args:
        qobj_dict: the dict of the PulseQobj to compress

    Returns:
        A compressed dict of the qobj
    """
    # In-place RLE pulse library for compression
    for pulse in qobj_dict["config"]["pulse_library"]:
        pulse["samples"] = _iqx_rle(pulse["samples"])

    return qobj_dict


def _iqx_rle(seq: List[Any]) -> List[Union[Tuple[Any], Tuple[Any, int]]]:
    """Run-length encodes a sequence.

    Constant subsequences are stored as a single term and count, rather than as the original subsequence.
    Counts equal to 1 are omitted.

    Args:
        seq: the sequence to encode

    Returns:
        A list of tuples that represent the encoding for the sequence
    """
    seq = [(k, sum(1 for _ in g)) for k, g in groupby(seq)]
    return [(c, rep) if rep > 1 else (c,) for c, rep in seq]



# 2023-05-22 following:
# https://learn.qiskit.org/course/ch-applications/hybrid-quantum-classical-neural-networks-with-pytorch-and-qiskit
import threading
import typing
from time import sleep

import numpy as np

# working with qiskit-aer 0.8.2
import qiskit
from qiskit import transpile

try:
    from tergite import Provider
    from tergite.types.backend import TergiteBackend
except:
    print("Could not import Tergite!")
from qiskit.result import marginal_counts
from qiskit_aer import AerSimulator
from utils import Utils

# Import Tergite
# from qiskit.providers.tergite import Tergite


class HybridLayerCircuit:
    """
    This class provides a simple interface for interaction
    with the quantum circuit
    """

    def __init__(
        self,
        backend: qiskit.providers.Backend = qiskit.Aer.get_backend("aer_simulator"),
        shots: int = 100,
        qubits: typing.List[int] = None,
        run_id: str = "default",
    ):
        self.backend = backend
        self.shots = shots
        self.qubits = qubits
        self.run_id = run_id

    def run(self, thetas: typing.Union[typing.List[float], np.ndarray]) -> np.ndarray:
        """
        Runs a circuit that looks like this:
        |0> ---|H|---|RY(theta_0)|---|M|
        |0> ---|H|---|RY(theta_1)|---|M|
        ...
        |0> ---|H|---|RY(theta_n)|---|M|

        Args:
            thetas: List of thetas for the RY gates

        Returns:
            The expectation value for the standard measurement in the computational basis

        """
        # TODO: here one could add more qubit bucketing options
        # n_qubits = 1 if self.qubits is None else len(self.qubits)
        qubits = [0] if self.qubits is None else self.qubits

        min_circuit_size = max(qubits) + 1

        Utils.append_thetas(self.run_id, thetas)

        circuit = qiskit.QuantumCircuit(min_circuit_size)
        circuit.h(qubits)
        for theta, i in list(zip(thetas, qubits)):
            circuit.ry(theta, i)
        circuit.measure_all()
        circuit.delay(100_000)
        t_qc = transpile(circuit, self.backend)

        job_args = tuple([t_qc])

        if isinstance(self.backend, AerSimulator):
            job_kwargs = {"shots": self.shots}
        elif isinstance(self.backend, TergiteBackend):
            job_kwargs = {"meas_level": 2, "meas_return": "single"}
        else:
            raise NotImplementedError("Cannot resolve backend!")

        job = self.backend.run(*job_args, **job_kwargs)

        waiting = True
        while waiting:
            try:
                job.result().get_counts()
                waiting = False
            except:
                sleep(0.25)

        expectations = []
        for i in qubits:
            marginalised_result = marginal_counts(job.result(), indices=[i])
            result = marginalised_result.get_counts()

            counts = np.array(list(result.values()))
            states = np.array(list(result.keys())).astype(np.float32)

            probabilities = counts / self.shots
            expectations.append(np.sum(states * probabilities))

        # TODO: here, one must as well think of how to average the expectation values by then later
        averaged_expectations = np.mean(np.array(expectations))
        return np.array([averaged_expectations], dtype=np.float32)



#!/usr/bin/env python3.9
# coding: utf-8

import contextlib
import time

import matplotlib.pyplot as plt
import numpy as np
import qiskit.circuit as circuit
import qiskit.compiler as compiler

with contextlib.redirect_stderr(None):
    from qiskit.ignis.verification.tomography import StateTomographyFitter
    from qiskit.ignis.verification.tomography import state_tomography_circuits

from datetime import datetime
from os import listdir, makedirs
from pathlib import Path
from shutil import move

from qiskit.providers.jobstatus import JobStatus
from qiskit.visualization import plot_bloch_multivector
from tqdm.auto import tqdm

from tergite.qiskit.providers import Tergite

folder = Path("demo_bloch_frames").resolve()


def save_old_frames():
    global folder
    saved_folder = folder / "saved_animations"
    makedirs(saved_folder, exist_ok=True)

    old_frames = [f for f in listdir(folder) if f.endswith(".jpg")]

    if len(old_frames):
        now_time = datetime.now()
        mstr = now_time.strftime("%Y%m%d%H%M%S")
        new_dir = saved_folder / mstr
        makedirs(new_dir, exist_ok=True)
        for f in old_frames:
            move(folder / f, new_dir / f)


save_old_frames()

# ----------------------------------------------------------------------------

chalmers = Tergite.get_provider()
backend = chalmers.get_backend("Nov7")
backend.set_options(shots=2000)

print(f"Loaded backend {backend.name} (QAL 9000)")

thetadef = -1 * np.asarray([0, np.pi / 2, np.pi])


def tomog_circs(theta):
    q = circuit.QuantumRegister(1)
    circ = circuit.QuantumCircuit(q)
    circ.barrier([0])
    circ.reset([0])

    circ.rx(theta, q[0])

    return state_tomography_circuits(circ, [q[0]])


print("Transpiling...")
with contextlib.redirect_stderr(None):
    precomputed_tomog_circs = [
        compiler.transpile(tomog_circs(theta), backend=backend) for theta in thetadef
    ]


def compute_new_frame(j: int):
    job = backend.run(precomputed_tomog_circs[j], meas_level=2, meas_return="single")
    while job.status() != JobStatus.DONE:
        time.sleep(1)  # blocking wait

    # fit state vector when result is ready
    fitter = StateTomographyFitter(job.result(), precomputed_tomog_circs[j])

    density_matrix = fitter.fit(method="lstsq")

    # compute frame and save to main memory
    _tmp = plot_bloch_multivector(
        density_matrix, reverse_bits=True, filename=folder / f"frame{j}.jpg"
    )
    plt.close(_tmp)  # close returned figure


# progress bar
for j in tqdm(range(len(thetadef)), desc="Reconstructing qubit state"):
    compute_new_frame(j)



import json
import time
from pprint import pprint

import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
import qiskit.circuit as circuit
import qiskit.compiler as compiler
import qiskit.pulse as pulse
import qiskit.quantum_info as qi
import retworkx.visualization as rxv
from matplotlib import pyplot as plt
from qiskit import Aer
from qiskit.ignis.verification.tomography import (
    StateTomographyFitter,
    state_tomography_circuits,
)
from qiskit.providers.jobstatus import JobStatus
from qiskit.providers.tergite import Tergite
from qiskit.visualization import (
    plot_bloch_multivector,
    plot_histogram,
    plot_state_city,
    plot_state_hinton,
    plot_state_paulivec,
    plot_state_qsphere,
)
from qiskit.visualization.pulse_v2.stylesheet import IQXDebugging, IQXSimple
from qiskit_experiments.library.tomography import StateTomography

backend = Aer.get_backend("aer_simulator")

fig, ax = plt.subplots()

im = ax.imshow(np.zeros((320, 1600)))
ax.axis("off")

thetadef = np.linspace(0, 1.09 * np.pi / 2, 20)  # overshoot by 9%


def tomog_circs(theta):
    q = circuit.QuantumRegister(5)
    circ = circuit.QuantumCircuit(q)
    circ.barrier()
    [circ.reset(qb) for qb in q]  # reset qubits

    # Rotate qubit 4 around the y-axis
    circ.ry(theta, q[4])

    # Rotate qubit 3 around the z-axis, spinning at np.pi/4 around x
    circ.rx(np.pi / 4, q[3])
    circ.rz(theta, q[3])

    # Rotate qubit 2 around the x-axis
    circ.rx(theta, q[2])

    # Rotate qubit 3 around the z-axis in the other direction, spinning at np.pi/4 around x
    circ.rx(-np.pi / 4, q[1])
    circ.rz(-theta, q[1])

    # Rotate qubit 0 around the y-axis in the other direction
    circ.ry(-theta, q[0])

    return state_tomography_circuits(circ, q)


precomputed_tomog_circs = [tomog_circs(theta) for theta in thetadef]


# function to update figure
def updatefig(j):
    job = backend.run(precomputed_tomog_circs[j], meas_level=2)
    while job.status() != JobStatus.DONE:
        time.sleep(0.1)  # blocking wait

    # fit state vector
    fitter = StateTomographyFitter(job.result(), precomputed_tomog_circs[j])
    density_matrix = fitter.fit(method="lstsq")

    # plot and save to main memory
    _tmp = plot_bloch_multivector(
        density_matrix, reverse_bits=True, filename=f"frames/frame{j}.jpg"
    )
    plt.close(_tmp)

    # load file (workaround, because cannot access the axis directly)
    image_j = plt.imread(f"frames/frame{j}.jpg")

    # show next frame
    im.set_array(image_j)
    return [im]


# start animation
ani = animation.FuncAnimation(
    fig, updatefig, frames=range(20), interval=50, blit=True, repeat=False
)
figManager = plt.get_current_fig_manager()
figManager.window.showMaximized()
plt.show()



"""Fixtures for end-to-end tests"""

import os
import re

import pytest
import requests
import requests_mock
from urllib3 import HTTPResponse

from tergite import Provider, Tergite
from tergite.services.configs import AccountInfo
from tests.utils.env import is_end_to_end, is_in_docker


@pytest.mark.skipif(not is_end_to_end(), reason="is end-to-end test")
@pytest.fixture
def mock_backend_servers():
    with requests_mock.Mocker(real_http=True) as mocker:
        if not is_in_docker():
            # use the network's url if running in docker, else proxy it to the host loopback url
            matcher = re.compile(r"http://qiskit_pulse_")
            mocker.register_uri(requests_mock.ANY, matcher, raw=proxy_request)
        yield mocker


@pytest.mark.skipif(not is_end_to_end(), reason="is end-to-end test")
@pytest.fixture
def backend_provider(mock_backend_servers) -> Provider:
    """A provider that is authenticated using API token"""
    api_url = os.environ.get("API_URL")
    api_token = os.environ.get("API_TOKEN")
    return Tergite.use_provider_account(
        service_name="test", url=api_url, token=api_token
    )


def proxy_request(request: requests.Request, context) -> HTTPResponse:
    """A proxy server for the backends

    This is essential as the backend urls contain domain names
    that are specific to the docker network

    Args:
        request: the received request
        context: the context for the mocker

    Returns:
        the response
    """
    url: str = request.url
    url = url.replace("qiskit_pulse_1q:8000", "localhost:8001")
    url = url.replace("qiskit_pulse_2q:8000", "localhost:8000")

    forwarded_response = requests.request(
        method=request.method,
        url=url,
        headers=request.headers,
        data=request.body,
    )
    context.status_code = forwarded_response.status_code

    return forwarded_response.raw



from datetime import datetime

import torch
from data_handling import IQDataGenerator
from learning import Net, eval_model, train_model
from torch.utils.data import DataLoader
from utils import Utils

# Set this global ID to some value or the model value that you want to learn
RUN_ID = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

# TODO: how many qubits do we wanna use?
QUBITS = [i for i in range(9)]
N_QUBITS = len(QUBITS)
SHOTS = 1024

TRAIN_EPOCHS = 10

N_TRAIN_SAMPLES = 100
N_EVAL_SAMPLES = 20

# TODO: which backend do we wanna use?
BACKEND = Utils.get_backend("new_nine_qubits", shots=SHOTS)

# generate data
train_loader = DataLoader(
    IQDataGenerator.get_2d_binary_class(n_samples=N_TRAIN_SAMPLES),
    batch_size=1,
    shuffle=True,
)
eval_loader = DataLoader(
    IQDataGenerator.get_2d_binary_class(n_samples=N_EVAL_SAMPLES),
    batch_size=1,
    shuffle=True,
)

# init model
model = Net(RUN_ID, n_qubits=N_QUBITS, qubits=QUBITS, backend=BACKEND, shots=SHOTS)

# train model
trained_model, losses = train_model(model, train_loader, TRAIN_EPOCHS)

# save
torch.save(trained_model, f"temp/{RUN_ID}.model")

accuracy = eval_model(trained_model, eval_loader)



import os
import pickle
import typing
from typing import TYPE_CHECKING

import numpy as np
import qiskit

if TYPE_CHECKING:
    from learning import Net

try:
    from tergite import Tergite
except:
    print("Could not import Tergite!")


class Utils:
    @staticmethod
    def get_backend(name: str, shots: int = 1024):
        if name == "AerSimulator":
            return qiskit.Aer.get_backend("aer_simulator")
        elif name is None:
            raise RuntimeError("Calling backend without providing name")
        else:
            tergite_provider = Tergite.get_provider()
            backend = tergite_provider.get_backend(name)
            backend.set_options(shots=shots)
            return backend

    @staticmethod
    def change_backend_layer(
        model: "Net", backend, qubits: typing.List[int], shots: int = 1024
    ) -> "Net":
        model.hybrid.quantum_circuit.backend = backend
        model.hybrid.quantum_circuit.shots = shots
        model.hybrid.quantum_circuit.qubits = qubits
        return model

    @staticmethod
    def append_thetas(run_id: str, thetas: np.ndarray):
        if not os.path.exists("temp"):
            os.makedirs("temp")
        filename = f"temp/{run_id}.thetas"
        stored_thetas = np.array([])
        if f"{run_id}.thetas" in os.listdir("temp"):
            stored_thetas = pickle.load(open(filename, "rb"))
        stored_thetas = np.concatenate((stored_thetas, np.array(thetas)))
        pickle.dump(stored_thetas, open(filename, "wb"))

    @staticmethod
    def load_thetas(run_id: str):
        filename = f"temp/{run_id}.thetas"
        return pickle.load(open(filename, "rb"))



# This code is part of Tergite
#
# (C) Copyright Chalmers Next Labs 2025
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Test utilities for the environment"""
import os


def is_end_to_end() -> bool:
    """Checks if current environment is end to end

    Returns:
        True if the environment is in END_TO_END settings, False otherwise
    """
    return os.environ.get("IS_END_TO_END", "False").lower() == "true"


def is_in_docker() -> bool:
    """Checks if this is running in docker

    Returns:
        True if running in docker container settings, False otherwise
    """
    return os.environ.get("IS_IN_DOCKER", "False").lower() == "true"



# This code is part of Tergite
#
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""tests for the get_backend method on tergite backend"""
import pytest
from qiskit.providers import QiskitBackendNotFoundError

from tergite import OpenPulseBackend, Provider, Tergite
from tergite.services.api_client.dtos import TergiteBackendConfig
from tests.utils.records import get_record

from ..utils.env import is_end_to_end
from .conftest import (
    API_TOKEN,
    API_URL,
    BACKENDS_LIST,
    GOOD_BACKENDS,
    INVALID_API_TOKENS,
    MALFORMED_BACKEND,
)

_INVALID_PARAMS = [
    (token, backend) for backend in GOOD_BACKENDS for token in INVALID_API_TOKENS
]


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_get_backend(api, backend_name):
    """Retrieves the right backend"""
    provider = _get_test_provider(url=API_URL)
    expected_json = get_record(BACKENDS_LIST, _filter={"name": backend_name})
    expected = OpenPulseBackend(
        data=TergiteBackendConfig(**expected_json), provider=provider, base_url=API_URL
    )
    got = provider.get_backend(backend_name)
    assert got == expected


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
def test_get_malformed_backend(api):
    """Raises TypeError if a malformed backend is returned"""
    provider = _get_test_provider(url=API_URL)

    with pytest.raises(
        QiskitBackendNotFoundError, match="No backend matches the criteria"
    ):
        provider.get_backend(MALFORMED_BACKEND)


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("backend_name", GOOD_BACKENDS)
def test_bearer_auth(bearer_auth_api, backend_name):
    """Retrieves the data if backend is shielded with basic auth"""
    provider = _get_test_provider(url=API_URL, token=API_TOKEN)
    expected_json = get_record(BACKENDS_LIST, _filter={"name": backend_name})
    expected = OpenPulseBackend(
        data=TergiteBackendConfig(**expected_json), provider=provider, base_url=API_URL
    )
    got = provider.get_backend(backend_name)
    assert got == expected


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("token, backend", _INVALID_PARAMS)
def test_invalid_bearer_auth(token, backend, bearer_auth_api):
    """Invalid bearer auth raises RuntimeError if backend is shielded with bearer auth"""
    provider = _get_test_provider(url=API_URL, token=token)
    with pytest.raises(RuntimeError, match="Error retrieving backends: Unauthorized"):
        provider.get_backend(backend)


def _get_test_provider(url: str, token: str = None) -> Provider:
    """Retrieves a provider to be used for testing given the optional token"""
    return Tergite.use_provider_account(service_name="test", url=url, token=token)



# This code is part of Tergite
#
# (C) Copyright Miroslav Dobsicek 2020
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


# this file is meant to be run interactively for objects inspection
# python -i tests/job_on_tergite.py

from qiskit import *

from tergite.qiskit.providers import Tergite

provider = Tergite.get_provider()
backend = provider.get_backend("pingu")

qr = QuantumRegister(3)
qc = QuantumCircuit(qr)
# qc.cz(qr[1], qr[2])
qc.x(qr[1])
qc.h(qr[2])

mytr = transpile(qc, backend)
myqobj = assemble(mytr, backend)

job = execute(mytr, backend)

# plot_histogram(job.result().get_counts()).show()



"""End-to-end tests for job execution"""

import pytest
from qiskit import circuit, compiler

from tergite import Job, Provider
from tests.utils.env import is_end_to_end


@pytest.mark.skipif(not is_end_to_end(), reason="is end-to-end test")
def test_simulator_1_qubit_gate(backend_provider: Provider):
    """Can run a single qubit gate operation on simulator when auth on MSS and backend is ON"""
    backend = backend_provider.get_backend("qiskit_pulse_1q")
    backend.set_options(shots=1024)

    qc = circuit.QuantumCircuit(1, 1)
    qc.x(0)
    qc.h(0)
    qc.measure_all()

    tc = compiler.transpile(qc, backend=backend)
    job: Job = backend.run(tc, meas_level=2, meas_return="single")
    job.wait_for_final_state(timeout=100)
    result = job.result().get_counts()

    assert (
        450 <= result["00"] <= 640
    ), f"Expected {result['00']} to be between 450 and 640"
    assert (
        450 <= result["01"] <= 640
    ), f"Expected {result['01']} to be between 450 and 640"


@pytest.mark.skipif(not is_end_to_end(), reason="is end-to-end test")
def test_simulator_2_qubit_gate(backend_provider: Provider):
    """Can run a two-qubit gate operation on simulator when auth on MSS and backend is ON"""
    backend = backend_provider.get_backend("qiskit_pulse_2q")
    backend.set_options(shots=1024)

    qc = circuit.QuantumCircuit(2, 2)
    qc.h(0)
    qc.h(1)
    qc.cx(0, 1)
    qc.measure_all()

    tc = compiler.transpile(qc, backend=backend)
    job: Job = backend.run(tc, meas_level=2, meas_return="single")
    job.wait_for_final_state(timeout=300)
    result = job.result().get_counts()

    assert (
        450 <= result["0000"] <= 640
    ), f"Expected {result['0000']} to be between 450 and 640"
    assert (
        450 <= result["0011"] <= 640
    ), f"Expected {result['0011']} to be between 450 and 640"



# This code is part of Tergite
#
# (C) Copyright Martin Ahindura 2023
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
import io
import json
import re
from pathlib import Path
from tempfile import gettempdir
from typing import Any, Dict

import h5py
import pytest
import requests_mock as rq_mock
from requests import Request

from tergite.compat.qiskit.qobj.encoder import IQXJsonEncoder as PulseQobj_encoder
from tests.utils.env import is_end_to_end
from tests.utils.fixtures import load_json_fixture

API_URL = "https://api.tergite.example"
QUANTUM_COMPUTER_URL = "http://loke.tergite.example"
API_TOKEN = "some-token"
BACKENDS_LIST = load_json_fixture("many_backends.json")
_QOBJ_RESULTS = load_json_fixture("qobj_results.json")
TEST_JOB_ID = "test_job_id"
TEST_QOBJ_ID = "test_qobj_id"
NUMBER_OF_SHOTS = 100

_BACKENDS_URL = f"{API_URL}/devices/"
_JOBS_URL = f"{API_URL}/jobs/"
_TEST_JOB_URL = f"{API_URL}/jobs/{TEST_JOB_ID}"
_JOB_UPLOAD_URL = f"{QUANTUM_COMPUTER_URL}/jobs"
_TEST_JOB_CANCEL_URL = f"{QUANTUM_COMPUTER_URL}/jobs/{TEST_JOB_ID}/cancel"
_TEST_RESULTS_DOWNLOAD_PATH = f"{QUANTUM_COMPUTER_URL}/test_file.hdf5"
_TEST_JOB_RESPONSE = {"job_id": TEST_JOB_ID, "upload_url": _JOB_UPLOAD_URL}
_HALF_NUMBER_OF_SHOTS = int(NUMBER_OF_SHOTS / 2)
_TMP_RESULTS_PATH = Path(gettempdir()) / f"{TEST_JOB_ID}.hdf5"
_CALIBRATIONS_REGEX = re.compile(f"^{API_URL}/calibrations/([\w-]+)")
_JOBS_REGISTER_URL_REGEX = re.compile(f"^{API_URL}/jobs/")
_JOBS_UPLOAD_URL_REGEX = re.compile(f"^http://([\w-]+)\.tergite\.example/jobs")
_JOBS_URL_REGEX = re.compile(f"{API_URL}/jobs/([\w-]+)")
_JOBS_CANCEL_URL_REGEX = re.compile(
    f"^http://([\w-]+)\.tergite\.example/jobs/([\w-]+)/cancel"
)
_JOBS_LOGFILE_URL_REGEX = re.compile(
    r"^http://([\w-]+)\.tergite\.example/test_file.hdf5"
)

TEST_JOB_RESULTS = {
    "status": "successful",
    "download_url": _TEST_RESULTS_DOWNLOAD_PATH,
    "result": {
        "memory": [
            (["0x1"] * _HALF_NUMBER_OF_SHOTS) + (["0x0"] * _HALF_NUMBER_OF_SHOTS)
        ],
    },
}
_TEST_JOB = {
    **_TEST_JOB_RESPONSE,
    **TEST_JOB_RESULTS,
    "device": "DUMMY_DEVICE",  # a dummy device that we might be able to replace in mock handler
    "calibration_date": "DUMMY_CALIB_DATE",  # a dummy date that we might be able to replace in mock handler
}
RAW_TEST_JOB_RESULTS = json.dumps(TEST_JOB_RESULTS).encode("utf-8")
GOOD_BACKENDS = ["Well-formed", "loke", "qiskit_pulse_1q", "qiskit_pulse_2q"]
TWO_QUBIT_BACKENDS = ["Well-formed", "loke", "qiskit_pulse_2q"]
MALFORMED_BACKEND = "Malformed"
INVALID_API_TOKENS = ["foo", "bar", "mayo", "API_USERNAME"]

_CALIBRATIONS = load_json_fixture("calibrations.json")
TEST_CALIBRATIONS_MAP = {item["name"]: {**item} for item in _CALIBRATIONS}
TEST_QUANTUM_COMPUTER_URL_MAP = {
    backend: f"http://{backend}.tergite.example" for backend in GOOD_BACKENDS
}
TEST_LOGFILE_DOWNLOAD_MAP = {
    backend: f"http://{backend}.tergite.example/test_file.hdf5"
    for backend in GOOD_BACKENDS
}
TEST_JOBS_MAP = {
    backend: {
        "job_id": f"{TEST_JOB_ID}-{backend}",
        "upload_url": f"{TEST_QUANTUM_COMPUTER_URL_MAP[backend]}/jobs",
    }
    for backend in GOOD_BACKENDS
}
TEST_JOB_RESULTS_MAP = {
    f"{TEST_JOB_ID}-{backend}": {
        **TEST_JOB_RESULTS,
        "device": backend,
        "calibration_date": TEST_CALIBRATIONS_MAP[backend]["last_calibrated"],
        "download_url": TEST_LOGFILE_DOWNLOAD_MAP[backend],
    }
    for backend in GOOD_BACKENDS
}
TEST_QOBJ_RESULTS_MAP = {
    item["header"]["backend_name"].lower(): {**item} for item in _QOBJ_RESULTS
}


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.fixture
def api(requests_mock):
    """The mock api without auth"""
    requests_mock.get(
        _BACKENDS_URL,
        headers={},
        json={"data": BACKENDS_LIST, "skip": 0, "limit": None},
    )

    # job registration
    requests_mock.post(_JOBS_URL, headers={}, json=_TEST_JOB_RESPONSE)
    # job upload
    requests_mock.post(_JOB_UPLOAD_URL, headers={}, status_code=200)
    # job details
    requests_mock.get(_TEST_JOB_URL, headers={}, json=_TEST_JOB)
    # job cancellation
    requests_mock.post(_TEST_JOB_CANCEL_URL, headers={})
    # download file
    requests_mock.get(
        _TEST_RESULTS_DOWNLOAD_PATH, headers={}, content=RAW_TEST_JOB_RESULTS
    )
    requests_mock.get(_CALIBRATIONS_REGEX, headers={}, json=_mock_calibrations_handler)
    yield requests_mock


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.fixture
def bearer_auth_api(requests_mock):
    """The mock api with bearer auth"""
    request_headers = {"Authorization": f"Bearer {API_TOKEN}"}

    no_auth_matcher = _without_headers(request_headers)
    no_auth_json = {"detail": "Unauthorized"}
    requests_mock.get(
        _BACKENDS_URL,
        request_headers=request_headers,
        json={"data": BACKENDS_LIST, "skip": 0, "limit": None},
    )
    requests_mock.get(
        _BACKENDS_URL,
        status_code=401,
        additional_matcher=no_auth_matcher,
        json=no_auth_json,
    )

    # job registration
    requests_mock.post(
        _JOBS_URL, request_headers=request_headers, json=_TEST_JOB_RESPONSE
    )
    requests_mock.post(
        _JOBS_URL,
        status_code=401,
        additional_matcher=no_auth_matcher,
        json=no_auth_json,
    )

    # job upload
    requests_mock.post(
        _JOB_UPLOAD_URL, request_headers=request_headers, status_code=200
    )
    requests_mock.post(
        _JOB_UPLOAD_URL,
        status_code=401,
        additional_matcher=no_auth_matcher,
        json=no_auth_json,
    )

    # job details
    requests_mock.get(_TEST_JOB_URL, request_headers=request_headers, json=_TEST_JOB)
    requests_mock.get(
        _TEST_JOB_URL,
        status_code=401,
        additional_matcher=no_auth_matcher,
        json=no_auth_json,
    )

    # job cancellation
    requests_mock.post(_TEST_JOB_CANCEL_URL, headers=request_headers)
    requests_mock.post(
        _TEST_JOB_CANCEL_URL,
        status_code=401,
        additional_matcher=no_auth_matcher,
        json=no_auth_json,
    )

    # download file
    requests_mock.get(
        _TEST_RESULTS_DOWNLOAD_PATH,
        request_headers=request_headers,
        content=RAW_TEST_JOB_RESULTS,
    )
    requests_mock.get(
        _TEST_RESULTS_DOWNLOAD_PATH,
        status_code=401,
        additional_matcher=no_auth_matcher,
        json=no_auth_json,
    )

    # calibrations
    requests_mock.get(
        _CALIBRATIONS_REGEX,
        request_headers=request_headers,
        json=_mock_calibrations_handler,
    )
    requests_mock.get(
        _CALIBRATIONS_REGEX,
        status_code=401,
        additional_matcher=no_auth_matcher,
        json=no_auth_json,
    )
    yield requests_mock


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.fixture
def tmp_results_file():
    """The path to the tmp file where results are downloaded"""
    yield _TMP_RESULTS_PATH
    _TMP_RESULTS_PATH.unlink()


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.fixture
def mock_tergiterc() -> Path:
    """The mock tergite rc file path"""
    tergiterc_file = Path(gettempdir()) / ".qiskit" / "test_tergiterc"
    tergiterc_file.parent.mkdir(parents=True, exist_ok=True)

    with open(tergiterc_file, mode="w") as file:
        pass

    yield tergiterc_file
    tergiterc_file.unlink(missing_ok=True)


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.fixture
def api_with_logfile(requests_mock):
    """A mock api fixture for tests that need to use TEST_JOB_RESULTS_LOGFILE."""
    requests_mock.get(
        _BACKENDS_URL,
        headers={},
        json={"data": BACKENDS_LIST, "skip": 0, "limit": None},
    )

    # Job registration
    requests_mock.post(
        _JOBS_REGISTER_URL_REGEX, headers={}, json=_mock_job_registration_handler
    )
    # Job upload
    requests_mock.post(_JOBS_UPLOAD_URL_REGEX, headers={}, status_code=200)

    # Job details - use TEST_JOB_RESULTS_LOGFILE
    requests_mock.get(_JOBS_URL_REGEX, headers={}, json=_mock_job_details_handler)

    # Download file - use hdf5_content
    requests_mock.get(
        _JOBS_LOGFILE_URL_REGEX, headers={}, content=_mock_logfile_download_handler
    )

    # job cancellation
    requests_mock.post(
        _JOBS_CANCEL_URL_REGEX, headers={}, content=_mock_job_cancellation_handler
    )

    # calibration request
    requests_mock.get(_CALIBRATIONS_REGEX, json=_mock_calibrations_handler)
    yield requests_mock


def _without_headers(headers):
    """Matches requests that don't have the given headers"""

    def matcher(request):
        for k, v in headers.items():
            if request.headers.get(k) != v:
                return True

        return False

    return matcher


def _mock_calibrations_handler(request: Request, context: Any) -> Dict[str, Any]:
    """Mock API handler for v2/calibrations/{name} endpoint

    Args:
        request: the request caught
        context: the object with the extra context passed when creating mock e.g. headers

    Returns:
        the JSON data in dict form to be returned on the given endpoint
    """
    matcher = _CALIBRATIONS_REGEX.match(request.url)
    try:
        backend_name = matcher.group(1)
        data = TEST_CALIBRATIONS_MAP[backend_name]
        return {**data}
    except (AttributeError, KeyError):
        raise rq_mock.NoMockAddress(request)


def _mock_job_registration_handler(request: Request, context: Any) -> Dict[str, Any]:
    """Mock API handler for POST jobs/ MSS endpoint

    Args:
        request: the request caught
        context: the object with the extra context passed when creating mock e.g. headers

    Returns:
        the JSON data in dict form to be returned on the given endpoint
    """
    try:
        payload = request.json()
        device = payload["device"]
        data = TEST_JOBS_MAP[device]
        return {**data, **payload}
    except (AttributeError, KeyError):
        raise rq_mock.NoMockAddress(request)


def _mock_job_details_handler(request: Request, context: Any) -> Dict[str, Any]:
    """Mock API handler for GET jobs/{job_id} MSS endpoint

    Args:
        request: the request caught
        context: the object with the extra context passed when creating mock e.g. headers

    Returns:
        the JSON data in dict form to be returned on the given endpoint
    """
    matcher = _JOBS_URL_REGEX.match(request.url)
    try:
        job_id = matcher.group(1)
        data = TEST_JOB_RESULTS_MAP[job_id]
        return {**data, "job_id": job_id}
    except (AttributeError, KeyError):
        raise rq_mock.NoMockAddress(request)


def _mock_logfile_download_handler(request: Request, context: Any):
    """Mock API handler for the logfile download endpoint

    Args:
        request: the request caught
        context: the object with the extra context passed when creating mock e.g. headers

    Returns:
        the data to be returned on the given endpoint
    """
    matcher = _JOBS_LOGFILE_URL_REGEX.match(request.url)
    try:
        backend = matcher.group(1)
        qobj = {**TEST_QOBJ_RESULTS_MAP[backend.lower()]}
        hdf5_file = io.BytesIO()
        with h5py.File(hdf5_file, "w") as hdf:
            header_group = hdf.create_group("header")
            qobj_metadata_group = header_group.create_group("qobj_metadata")

            qobj_metadata_group.attrs["shots"] = qobj["config"]["shots"]
            qobj_metadata_group.attrs["qobj_id"] = qobj["qobj_id"]
            qobj_metadata_group.attrs["num_experiments"] = len(qobj["experiments"])

            qobj_data_group = header_group.create_group("qobj_data")
            experiment_data = json.dumps(qobj, cls=PulseQobj_encoder, indent="\t")
            qobj_data_group.attrs["experiment_data"] = experiment_data

        hdf5_file.seek(0)
        return hdf5_file.read()
    except (AttributeError, KeyError):
        raise rq_mock.NoMockAddress(request)


def _mock_job_cancellation_handler(request: Request, context: Any):
    """Mock API handler for the job cancellation endpoint

    Args:
        request: the request caught
        context: the object with the extra context passed when creating mock e.g. headers

    Returns:
        the JSON data in dict form to be returned on the given endpoint
    """
    matcher = _JOBS_CANCEL_URL_REGEX.match(request.url)
    try:
        backend = matcher.group(1)
        job_id = matcher.group(1)
        return None
    except (AttributeError, KeyError):
        raise rq_mock.NoMockAddress(request)



from pprint import pprint

import qiskit.quantum_info as qi
from qiskit import *
from qiskit.ignis.verification.tomography import (
    StateTomographyFitter,
    state_tomography_circuits,
)
from qiskit.visualization import (
    plot_bloch_multivector,
    plot_histogram,
    plot_state_city,
    plot_state_hinton,
    plot_state_paulivec,
    plot_state_qsphere,
)

from tergite.qiskit.providers import Tergite

#
# Note: This code has been used on "Quantum software" PoC hearing
# on January 25, 2022.
# Run with: python -i bell_state.py
#

# Backend ------------------------------------------------------

# Tergite
provider = Tergite.get_provider()
backend = provider.get_backend("pingu")

# Aer simulator
# backend = Aer.get_backend("aer_simulator")

# IBM
# account = IBMQ.load_account()
# backend= account.get_backend("ibmq_belem")


# Circuit ------------------------------------------------------
q = QuantumRegister(3)
circ = QuantumCircuit(q)
circ.h(q[0])
circ.cx(q[0], q[1])

print("Input circuit")
print(circ)

# State tomography circuits ------------------------------------
target_state = qi.Statevector.from_instruction(circ)
tomography_circuits = state_tomography_circuits(circ, [q[0], q[1], q[2]])

print("State tomography circuits")
print(tomography_circuits[0])
print("")
print(tomography_circuits[1])
print("")
print(tomography_circuits[2])


# Execute ------------------------------------------------------
job = execute(tomography_circuits, backend, shots=1000)

input("Executing job, press enter once it's finished")


# Fit  ---------------------------------------------------------
fitter = StateTomographyFitter(job.result(), tomography_circuits)
density_matrix = fitter.fit(method="lstsq")
traced_out = qi.partial_trace(density_matrix, [2])
plot_state_city(traced_out).show()

fidelity = qi.state_fidelity(density_matrix, target_state)
print("State Fidelity: F = {:.5f}".format(fidelity))



from datetime import datetime

import torch
import torch.nn as nn
from data_handling import IQDataGenerator
from learning import Net
from pylab import cos, sin
from qutip import basis
from torch import optim
from torch.utils.data import DataLoader
from tqdm import tqdm
from utils import Utils
from visuals import animate_bloch

RUN_ID = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
N_QUBITS = 9
TRAIN_EPOCHS = 5

N_TRAIN_SAMPLES = 100
N_EVAL_SAMPLES = 25

BACKEND = Utils.get_backend("AerSimulator")

# generate data
train_loader = DataLoader(
    IQDataGenerator.get_2d_binary_class(n_samples=N_TRAIN_SAMPLES),
    batch_size=1,
    shuffle=True,
)
eval_loader = DataLoader(
    IQDataGenerator.get_2d_binary_class(n_samples=N_EVAL_SAMPLES),
    batch_size=1,
    shuffle=True,
)

# init model
model = Net(RUN_ID, n_qubits=N_QUBITS, backend=BACKEND)
optimizer = optim.Adam(model.parameters(), lr=0.001)
loss_func = nn.NLLLoss()

loss_list = []

# train
model.train()
for epoch in range(TRAIN_EPOCHS):
    total_loss = []
    for batch_idx, (data, target) in tqdm(
        enumerate(train_loader), desc=f"{epoch = }", unit="samples"
    ):
        optimizer.zero_grad()
        output = model(data)
        loss = loss_func(output, target)
        loss.backward()
        optimizer.step()

        total_loss.append(loss.item())

    loss_list.append(sum(total_loss) / len(total_loss))

    print(
        "\nTraining [{:.0f}%]\tLoss: {:.4f}".format(
            100.0 * (epoch + 1) / TRAIN_EPOCHS, loss_list[-1]
        )
    )

# save
torch.save(model, f"{RUN_ID}_example.model")

# load
loaded_model = torch.load(f"{RUN_ID}_example.model")

# eval
loaded_model.eval()
with torch.no_grad():
    correct = 0
    for batch_idx, (data, target) in enumerate(eval_loader):
        output = loaded_model(data)

        pred = output.argmax(dim=1, keepdim=True)
        correct += pred.eq(target.view_as(pred)).sum().item()

        loss = loss_func(output, target)
        total_loss.append(loss.item())

    print(
        "Performance on test data:\n\tLoss: {:.4f}\n\tAccuracy: {:.1f}%".format(
            sum(total_loss) / len(total_loss), correct / len(eval_loader) * 100
        )
    )

states = []
thetas = Utils.load_thetas(RUN_ID)
print(len(thetas))
for theta in thetas:
    states.append((cos(theta / 2) * basis(2, 0) + sin(theta / 2) * basis(2, 1)).unit())

animate_bloch(states, duration=0.1, save_all=False, filename=f"{RUN_ID}_blochsphere")



import time

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import requests
from qiskit.providers.jobstatus import JobStatus
from qiskit.providers.tergite import Tergite

# Task: Estimate Pi numerically
# Area of circle = pi*r**2
#
# If we compute the area of a quarter of a circle and multiply it by 4, we should get Pi = 3.14159...
# The size of the circle doesn't matter, so let's take the top right quadrant of the unit circle.


def f(x):
    return np.sqrt(1 - x**2)


xdef = np.linspace(0, 1)

# ---------- Riemann estimate
riemann_est = sum(f(xdef) * np.diff(xdef)[0])
riemann_est_pi = riemann_est * 4


# ---------- Monte Carlo estimate, from pseudo random numbers
def mc_hitmiss(fn: callable, *, M: int) -> tuple:
    np.random.seed(0)
    U = np.random.rand(M, 2)
    estimate = np.mean(
        U[:, 1] <= fn(U[:, 0])
    )  # <- This is the estimator for integral value
    return estimate, U


mc_est, mc_samples = mc_hitmiss(f, M=400)
mc_est_pi = mc_est * 4

# ---------- Monte Carlo estimate, from Quantum Computer

mss_url = "http://qtl-elemos.mc2.chalmers.se:8002"
chalmers = Tergite.get_provider()
backend = chalmers.get_backend("Nov7")


# dummy class for testing, implement this before the demo
class _test:
    def job_id(self: object) -> str:
        return "7ef2ab6d-d33a-4849-82ca-2e183fc457ce"

    def status(self: object) -> JobStatus:
        if np.random.rand() < 0.2:
            return JobStatus.DONE
        else:
            return JobStatus.QUEUED


########################################################
job_ids = [
    "5a00247d-d22b-44ac-a00a-5d21c087f65d",
    "e904e367-b64b-4380-92eb-151f06ed5faf",
    "d044505c-4c92-4a34-8f7a-b880a03d8069",
]


def qmc_hitmiss(fn: callable, *, current_U: np.ndarray, job_id: str, M: int) -> tuple:
    #    job = _test()  # <- REPLACE THIS WITH SOMETHING THATS WAITABLE
    #    while job.status() != JobStatus.DONE:
    #        time.sleep(2)
    #    response = requests.get(mss_url + "/rng/" + str(job.job_id()))

    response = requests.get(mss_url + "/rng/" + job_id)
    N = 0
    if response.ok:
        data = response.json()
        N = (
            data["N"] // 2
        )  #  Half of array is X coordinates half is Y coordinates, N must be int
        X = np.asarray(data["numbers"][:N])  # first M are X coordinate
        Y = np.asarray(data["numbers"][N : (N + N)])  # second M are Y coordinate

    else:
        raise RuntimeError(
            f"Unable to fetch random numbers from MSS, response: {response}"
        )

    # these numbers are random 32-bit integers in the range [-2**31, 2**31 - 1]
    # so we needo to convert them to floats in the range [0,1]
    shape_size = current_U.shape[0]
    U = np.resize(current_U, (shape_size + N, 2))
    U[shape_size:, 0] = (X + 2**31) / (2**32)
    U[shape_size:, 1] = (Y + 2**31) / (2**32)

    estimate = np.mean(
        U[:, 1] <= fn(U[:, 0])
    )  # <- This is the estimator for integral value
    return estimate, U


# ---------- Plotting
text_settings = {"fontsize": 20}
plt.ion()
fig, axs = plt.subplots(1, 3, figsize=(17, 6.5), sharex=True, sharey=True)


def plot_estimate(ax: object, estimate: float):
    textstr = "$\hat\pi = {0:.5g}$".format(estimate)
    props = dict(boxstyle="round", facecolor="wheat", alpha=0.95)
    # place a text box in upper left in axes coords
    ax.text(
        0.15,
        0.15,
        textstr,
        transform=ax.transAxes,
        verticalalignment="top",
        bbox=props,
        fontsize=25,
    )


# plot riemann rectangles
riemann_rects = [
    mpl.patches.Rectangle(
        xy=(x, 0),
        width=np.diff(xdef)[0],
        height=f(x),
        ec="black",
        fc="none",
        alpha=0.75,
    )
    for x in xdef
]
for r in riemann_rects:
    axs[0].add_patch(r)
plot_estimate(axs[0], riemann_est_pi)

# plot monte carlo samples for the MC and QMC methods


axs[0].set_title("Riemann", **text_settings)
axs[1].set_title("Monte Carlo", **text_settings)
axs[2].set_title("Quantum Monte Carlo", **text_settings)

for ax in axs:
    ax.plot(np.linspace(0, 1, 1000), f(np.linspace(0, 1, 1000)), lw=3.5, zorder=99)
    ax.set_ylabel("f(x)", **text_settings)
    ax.set_xlabel("x", **text_settings)
    ax.grid(zorder=-99, alpha=0.3)
    ax.tick_params(axis="both", which="minor", labelsize=text_settings["fontsize"])
    ax.tick_params(axis="both", which="major", labelsize=text_settings["fontsize"])
fig.suptitle("$f(x) = (1-x^2)^{1/2}$", **text_settings)


def plot_samples(ax: object, samples: np.array):
    inside = np.sqrt(samples[:, 0] ** 2 + samples[:, 1] ** 2) < 1
    ax.scatter(samples[inside, 0], samples[inside, 1], color="C1")
    ax.scatter(samples[~inside, 0], samples[~inside, 1], color="C2")


plt.draw()
plot_samples(axs[1], mc_samples)
plot_estimate(axs[1], mc_est_pi)
x, y = [], []
qmc_in = axs[2].scatter(x, y, color="C1")
qmc_out = axs[2].scatter(x, y, color="C2")

plt.draw()

########## LOOP THROUGH  THE IDS
U = np.array([])
for current_job_id in job_ids:
    qmc_est, qmc_samples = qmc_hitmiss(f, current_U=U, job_id=current_job_id, M=400)
    U = qmc_samples
    print(U.shape)
    # print( qmc_samples )

    inside = np.sqrt(qmc_samples[:, 0] ** 2 + qmc_samples[:, 1] ** 2) < 1
    qmc_in.set_offsets(np.c_[qmc_samples[inside, 0], qmc_samples[inside, 1]])
    qmc_out.set_offsets(np.c_[qmc_samples[~inside, 0], qmc_samples[~inside, 1]])

    qmc_est_pi = qmc_est * 4
    textstr = "$\hat\pi = {0:.5g}$".format(qmc_est_pi)
    props = dict(boxstyle="round", facecolor="wheat", alpha=0.95)
    # place a text box in upper left in axes coords
    axs[2].text(
        0.15,
        0.15,
        textstr,
        transform=axs[2].transAxes,
        verticalalignment="top",
        bbox=props,
        fontsize=25,
    )

    fig.canvas.draw_idle()
    plt.pause(2)

    print(f"{ qmc_est_pi = }")
    # plot_samples(axs[2], qmc_samples)
    # plot_estimate(axs[2], qmc_est_pi)


# print(f"Riemann approximation of Pi {np.pi} from f(x) over [0,1] :", f"4*{riemann_qtr_pi} = ", 4*riemann_qtr_pi)


# fig.tight_layout()
# plt.show()



import re
from datetime import datetime

import pytest

from tergite.services.configs import AccountInfo
from tergite.types import provider_factory as factory
from tests.utils.env import is_end_to_end
from tests.utils.fixtures import load_json_fixture

_PROVIDER_ACCOUNTS = load_json_fixture("provider_accounts.json")


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("account_data", _PROVIDER_ACCOUNTS)
def test_use_provider_account(account_data, mock_tergiterc):
    """use_provider_account without save option returns the given provider account"""
    _tergite = factory.ProviderFactory(rc_file=mock_tergiterc)
    provider = _tergite.use_provider_account(**account_data)
    assert provider.account == AccountInfo(**account_data)


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("account_data", _PROVIDER_ACCOUNTS)
def test_use_provider_account_save(account_data, mock_tergiterc):
    """use_provider_account with save option saves the given provider account"""
    expected_config = (
        f"[service {account_data['service_name']}]\n" f"url = {account_data['url']}\n"
    )
    if "token" in account_data:
        expected_config += f"token = {account_data['token']}\n"

    extras = account_data.get("extras", {})
    for k, v in extras.items():
        expected_config += f"{k} = {v}\n"

    _tergite = factory.ProviderFactory(rc_file=mock_tergiterc)
    provider = _tergite.use_provider_account(**account_data, save=True)
    with open(mock_tergiterc, "r") as file:
        actual_config = file.read()

    assert expected_config in actual_config
    assert provider.account == AccountInfo(**account_data)


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("account_data", _PROVIDER_ACCOUNTS)
def test_get_provider(account_data, mock_tergiterc):
    """get_provider returns the provider with the given service name"""
    _tergite = factory.ProviderFactory(rc_file=mock_tergiterc)
    expected = _tergite.use_provider_account(**account_data, save=True)
    got = _tergite.get_provider(account_data["service_name"])
    assert got == expected


@pytest.mark.skipif(is_end_to_end(), reason="is not end-to-end test")
@pytest.mark.parametrize("account_data", _PROVIDER_ACCOUNTS)
def test_get_provider_non_existing(account_data, mock_tergiterc):
    """get_provider for non-existing service returns first provider"""
    _tergite = factory.ProviderFactory(rc_file=mock_tergiterc)
    _tergite.use_provider_account(**account_data, save=True)
    provider = _tergite.get_provider(datetime.now().isoformat())

    with open(mock_tergiterc, "r") as file:
        conf = file.read()

    first_service_name = re.search(r"\[service\s+(.+?)]\n", conf).group(1)
    assert provider.account.service_name == first_service_name
